# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Test Context.

The test `Context` object provides the context of the wider Juju system that the
specific `State` exists in, and the events that can be executed on that `State`.
"""

from __future__ import annotations

import functools
import pathlib
import tempfile
from collections.abc import Callable, Mapping
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Generic

from typing_extensions import deprecated

import ops
from ops._private.harness import ActionFailed

from ._runtime import Runtime
from .errors import (
    AlreadyEmittedError,
    ContextSetupError,
    MetadataNotFoundError,
)
from .logger import logger as scenario_logger
from .state import (
    CharmType,
    CheckInfo,
    Container,
    Notice,
    Secret,
    Storage,
    _Action,
    _CharmSpec,
    _Event,
)

if TYPE_CHECKING:  # pragma: no cover
    from opentelemetry.sdk.trace import ReadableSpan

    from ops._private.harness import ExecArgs

    from ._ops_main_mock import Ops
    from .state import (
        AnyJson,
        JujuLogLine,
        RelationBase,
        State,
        _EntityStatus,
    )

logger = scenario_logger.getChild('runtime')

_DEFAULT_JUJU_VERSION = '3.6.4'


class Manager(Generic[CharmType]):
    """Context manager to offer test code some runtime charm object introspection.

    This class should not be instantiated directly: use a :class:`Context`
    in a ``with`` statement instead, for example::

        ctx = Context(MyCharm)
        with ctx(ctx.on.start(), State()) as manager:
            manager.charm.setup()
            manager.run()
    """

    def __init__(
        self,
        ctx: Context[CharmType],
        arg: _Event,
        state_in: State,
    ):
        self._ctx = ctx
        self._arg = arg
        self._state_in = state_in

        self._emitted: bool = False

        self.ops: Ops[CharmType] | None = None

    @property
    def charm(self) -> CharmType:
        """The charm object instantiated by ops to handle the event.

        The charm is only available during the context manager scope.
        """
        if self.ops is None or self.ops.charm is None:
            raise RuntimeError(
                'you should __enter__ this context manager before accessing this',
            )
        return self.ops.charm

    @property
    def _runner(self):
        return self._ctx._run

    def __enter__(self):
        self._wrapped_ctx = wrapped_ctx = self._runner(self._arg, self._state_in)
        self.ops = wrapped_ctx.__enter__()
        return self

    def run(self) -> State:
        """Emit the event and proceed with charm execution.

        This can only be done once.
        """
        if self._emitted:
            raise AlreadyEmittedError('Can only run once.')
        if not self.ops:
            raise RuntimeError(
                'you should __enter__ this context manager before running it',
            )
        self._emitted = True
        self.ops.run()

        # wrap up Runtime.exec() so that we can gather the output state
        self._wrapped_ctx.__exit__(None, None, None)

        assert self._ctx._output_state is not None
        return self._ctx._output_state

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        if not self._emitted:
            logger.debug(
                "user didn't emit the event within the context manager scope. "
                'Doing so implicitly upon exit...',
            )
            self.run()
        if exc_type is not None:
            self._wrapped_ctx.__exit__(exc_type, exc_val, exc_tb)


def _copy_doc(original_func: Callable[..., Any]):
    """Copy the docstring from `original_func` to the wrapped function."""

    def decorator(wrapper_func: Callable[..., Any]):
        @functools.wraps(wrapper_func)
        def wrapped(*args: Any, **kwargs: Any):
            return wrapper_func(*args, **kwargs)

        wrapped.__doc__ = original_func.__doc__
        return wrapped

    return decorator


class CharmEvents:
    """Events generated by Juju or ops pertaining to the application lifecycle.

    The events listed as attributes of this class should be accessed via the
    :attr:`Context.on` attribute. For example::

        ctx.run(ctx.on.config_changed(), state)

    This behaves similarly to the :class:`ops.CharmEvents` class but is much
    simpler as there are no dynamically named attributes, and no ``__getattr__``
    version to get events. In addition, all of the attributes are methods,
    which are used to connect the event to the specific object that they relate
    to (or, for simpler events like "start" or "stop", take no arguments).
    """

    @staticmethod
    @_copy_doc(ops.InstallEvent)
    def install():  # noqa: D102
        return _Event('install')

    @staticmethod
    @_copy_doc(ops.StartEvent)
    def start():  # noqa: D102
        return _Event('start')

    @staticmethod
    @_copy_doc(ops.StopEvent)
    def stop():  # noqa: D102
        return _Event('stop')

    @staticmethod
    @_copy_doc(ops.RemoveEvent)
    def remove():  # noqa: D102
        return _Event('remove')

    @staticmethod
    @_copy_doc(ops.UpdateStatusEvent)
    def update_status():  # noqa: D102
        return _Event('update_status')

    @staticmethod
    @_copy_doc(ops.ConfigChangedEvent)
    def config_changed():  # noqa: D102
        return _Event('config_changed')

    @staticmethod
    @_copy_doc(ops.UpgradeCharmEvent)
    def upgrade_charm():  # noqa: D102
        return _Event('upgrade_charm')

    @staticmethod
    @_copy_doc(ops.PreSeriesUpgradeEvent)
    def pre_series_upgrade():  # noqa: D102
        return _Event('pre_series_upgrade')

    @staticmethod
    @_copy_doc(ops.PostSeriesUpgradeEvent)
    def post_series_upgrade():  # noqa: D102
        return _Event('post_series_upgrade')

    @staticmethod
    @_copy_doc(ops.LeaderElectedEvent)
    def leader_elected():  # noqa: D102
        return _Event('leader_elected')

    @staticmethod
    @_copy_doc(ops.SecretChangedEvent)
    def secret_changed(secret: Secret):  # noqa: D102
        if secret.owner:
            raise ValueError(
                'This unit will never receive secret-changed for a secret it owns.',
            )
        return _Event('secret_changed', secret=secret)

    @staticmethod
    @_copy_doc(ops.SecretExpiredEvent)
    def secret_expired(secret: Secret, *, revision: int):  # noqa: D102
        if not secret.owner:
            raise ValueError(
                'This unit will never receive secret-expire for a secret it does not own.',
            )
        return _Event('secret_expired', secret=secret, secret_revision=revision)

    @staticmethod
    @_copy_doc(ops.SecretRotateEvent)
    def secret_rotate(secret: Secret):  # noqa: D102
        if not secret.owner:
            raise ValueError(
                'This unit will never receive secret-rotate for a secret it does not own.',
            )
        return _Event('secret_rotate', secret=secret)

    @staticmethod
    @_copy_doc(ops.SecretRemoveEvent)
    def secret_remove(secret: Secret, *, revision: int):  # noqa: D102
        if not secret.owner:
            raise ValueError(
                'This unit will never receive secret-removed for a secret it does not own.',
            )
        return _Event('secret_remove', secret=secret, secret_revision=revision)

    @staticmethod
    def collect_app_status():
        """Event triggered at the end of every hook to collect app statuses for evaluation."""
        return _Event('collect_app_status')

    @staticmethod
    def collect_unit_status():
        """Event triggered at the end of every hook to collect unit statuses for evaluation."""
        return _Event('collect_unit_status')

    @staticmethod
    @_copy_doc(ops.RelationCreatedEvent)
    def relation_created(relation: RelationBase):  # noqa: D102
        return _Event(f'{relation.endpoint}_relation_created', relation=relation)

    @staticmethod
    @_copy_doc(ops.RelationJoinedEvent)
    def relation_joined(relation: RelationBase, *, remote_unit: int | None = None):  # noqa: D102
        return _Event(
            f'{relation.endpoint}_relation_joined',
            relation=relation,
            relation_remote_unit_id=remote_unit,
        )

    @staticmethod
    @_copy_doc(ops.RelationChangedEvent)
    def relation_changed(  # noqa: D102
        relation: RelationBase,
        *,
        remote_unit: int | None = None,
    ):
        return _Event(
            f'{relation.endpoint}_relation_changed',
            relation=relation,
            relation_remote_unit_id=remote_unit,
        )

    @staticmethod
    @_copy_doc(ops.RelationDepartedEvent)
    def relation_departed(  # noqa: D102
        relation: RelationBase,
        *,
        remote_unit: int | None = None,
        departing_unit: int | None = None,
    ):
        return _Event(
            f'{relation.endpoint}_relation_departed',
            relation=relation,
            relation_remote_unit_id=remote_unit,
            relation_departed_unit_id=departing_unit,
        )

    @staticmethod
    @_copy_doc(ops.RelationBrokenEvent)
    def relation_broken(relation: RelationBase):  # noqa: D102
        return _Event(f'{relation.endpoint}_relation_broken', relation=relation)

    @staticmethod
    @_copy_doc(ops.StorageAttachedEvent)
    def storage_attached(storage: Storage):  # noqa: D102
        return _Event(f'{storage.name}_storage_attached', storage=storage)

    @staticmethod
    @_copy_doc(ops.StorageDetachingEvent)
    def storage_detaching(storage: Storage):  # noqa: D102
        return _Event(f'{storage.name}_storage_detaching', storage=storage)

    @staticmethod
    @_copy_doc(ops.PebbleReadyEvent)
    def pebble_ready(container: Container):  # noqa: D102
        return _Event(f'{container.name}_pebble_ready', container=container)

    @staticmethod
    @_copy_doc(ops.PebbleCustomNoticeEvent)
    def pebble_custom_notice(container: Container, notice: Notice):  # noqa: D102
        return _Event(
            f'{container.name}_pebble_custom_notice',
            container=container,
            notice=notice,
        )

    @staticmethod
    @_copy_doc(ops.PebbleCheckFailedEvent)
    def pebble_check_failed(container: Container, info: CheckInfo):  # noqa: D102
        return _Event(
            f'{container.name}_pebble_check_failed',
            container=container,
            check_info=info,
        )

    @staticmethod
    @_copy_doc(ops.PebbleCheckRecoveredEvent)
    def pebble_check_recovered(container: Container, info: CheckInfo):  # noqa: D102
        return _Event(
            f'{container.name}_pebble_check_recovered',
            container=container,
            check_info=info,
        )

    @staticmethod
    @_copy_doc(ops.ActionEvent)
    def action(  # noqa: D102
        name: str,
        params: Mapping[str, AnyJson] | None = None,
        id: str | None = None,
    ):
        kwargs: dict[str, Any] = {}
        if params:
            kwargs['params'] = params
        if id:
            kwargs['id'] = id
        return _Event(f'{name}_action', action=_Action(name, **kwargs))

    @staticmethod
    def custom(
        event: ops.BoundEvent,
        *args: Any,
        **kwargs: Any,
    ):
        """Event triggered by a charm library.

        For example, suppose that a library uses a ``DatabaseRequirer`` object
        to emit a ``DatabaseReadyEvent`` event, with charm code::

            class MyCharm(ops.CharmBase):
                def __init__(self, framework: ops.Framework):
                    super().__init__(framework)
                    self.db = DatabaseRequirer(self, 'db-relation')
                    framework.observe(self.db.on.ready, self._on_db_ready)

        To emit a ``DatabaseReadyEvent`` event::

            ctx.run(ctx.on.custom(
                DatabaseRequirer.on.ready, 1, 2, foo='bar'
            ), state)

        Custom events do not have access to the Juju context of the originating
        hook event.

        If the custom event takes more arguments than an :attr:`ops.Handle`,
        pass these as additional arguments to ``custom()``. Any of these
        arguments that are State components (such as
        :attr:`ops.testing.Relation`) will be converted to their ops
        counterparts (such as :attr:`ops.Relation`).
        """
        if issubclass(event.event_type, ops.HookEvent):
            raise ValueError(
                'custom events should subclass EventBase directly - testing events that inherit '
                'from a Juju hook event must be done by running the underlying Juju event, for '
                'example: ctx.run(ctx.on.relation_changed(relation), state)'
            )
        return _Event(
            f'{{custom}}.{type(event.emitter).__name__}.{event.event_kind}',
            custom_event=event,
            custom_event_args=args,
            custom_event_kwargs=kwargs,
        )


class Context(Generic[CharmType]):
    """Testing context in which events are emitted.

    The ``Context`` represents a charm's execution context, simulated for the
    test execution. It is the main entry point to running a test, and contains:

    - the charm source code being executed
    - the metadata files associated with it
    - a charm project repository root
    - metadata about the Juju environment being simulated (version, unit name, and so on)

    After you have instantiated ``Context``, typically you will call :meth:`run()` to execute the
    charm once, write any assertions you like on the output state returned by the call, write any
    assertions you like on the ``Context`` attributes, then discard the ``Context``.

    Each ``Context`` instance is in principle designed to be single-use:
    ``Context`` is not cleaned up automatically between charm runs.

    Any side effects generated by executing the charm, that are not part of the mocked Juju
    ``State``, are also stored in the ``Context``:

    - :attr:`juju_log`
    - :attr:`app_status_history`
    - :attr:`unit_status_history`
    - :attr:`workload_version_history`
    - :attr:`removed_secret_revisions`
    - :attr:`requested_storages`
    - :attr:`emitted_events`
    - :attr:`action_logs`
    - :attr:`action_results`
    - :attr:`trace_data`

    This allows you to write assertions not only on the output state, but also, to some
    extent, on the path the charm took to get there.

    A typical test will look like::

        from charm import MyCharm, MyCustomEvent  # noqa

        def test_foo():
            # Arrange: set the context up
            ctx = Context(MyCharm)
            # Act: prepare the state and emit an event
            state_out = ctx.run(ctx.on.update_status(), State())
            # Assert: verify the output state is what you think it should be
            assert state_out.unit_status == ActiveStatus('foobar')
            # Assert: verify the Context contains what you think it should
            assert len(ctx.emitted_events) == 4
            assert isinstance(ctx.emitted_events[3], MyCustomEvent)

    If you need access to the charm object that will handle the event, use the
    class in a ``with`` statement, like::

        def test_foo():
            ctx = Context(MyCharm)
            with ctx(ctx.on.start(), State()) as mgr:
                mgr.charm._some_private_setup()
                mgr.run()
                assert mgr.charm._some_private_attribute == "bar"

    Note that you can't call ``run()`` multiple times. The context 'pauses' ops
    right before emitting the event, but otherwise this is a regular test; you
    can't emit multiple events in a single charm execution.
    """

    app_name: str
    """The name of the application that this charm is deployed as."""

    unit_id: int
    """The unit ID that this charm is deployed to."""

    juju_version: str
    """The Juju model version being simulated."""

    juju_log: list[JujuLogLine]
    """A record of what the charm has sent to juju-log"""

    app_trusted: bool
    """Whether the charm has Juju trust.

    Juju trust means deployed with ``--trust`` or added with ``juju trust``.
    """

    app_status_history: list[_EntityStatus]
    """A record of the app statuses the charm has set.

    Assert that the charm has followed the expected path by checking the app
    status history like so::

        ctx = Context(MyCharm)
        state_out = ctx.run(ctx.on.start(), State())
        assert ctx.app_status_history == [
            UnknownStatus(),
            MaintenanceStatus('...'),
        ]
        assert state_out.app_status == ActiveStatus()

    Note that the *current* status is **not** in the app status history.
    """

    unit_status_history: list[_EntityStatus]
    """A record of the unit statuses the charm has set.

    Assert that the charm has followed the expected path by checking the unit
    status history like so::

        ctx = Context(MyCharm)
        state_out = ctx.run(ctx.on.start(), State())
        assert ctx.unit_status_history == [
            UnknownStatus(),
            MaintenanceStatus('...'),
            WaitingStatus('...'),
        ]
        assert state_out.unit_status == ActiveStatus()

    Note that the *current* status is **not** in the unit status history.

    Also note that, unless you initialise the State with a preexisting status,
    the first status in the history will always be ``unknown``.
    """

    exec_history: dict[str, list[ExecArgs]]
    """A record of the exec calls made by the charm."""

    workload_version_history: list[str]
    """A record of the workload versions the charm has set.

    Assert that the charm has set one or more workload versions during a hook
    execution::

        ctx = Context(MyCharm)
        ctx.run(ctx.on.start(), State())
        assert ctx.workload_version_history == ['1', '1.2', '1.5']

    Note that the *current* version is **not** in the version history.
    """

    removed_secret_revisions: list[int]
    """A record of the secret revisions the charm has removed"""

    emitted_events: list[ops.EventBase]
    """A record of the events (including custom) that the charm has processed.

    You can configure which events will be captured by passing the following
    arguments to ``Context``:

    -  `capture_deferred_events`: If you want to include re-emitted deferred events.
    -  `capture_framework_events`: If you want to include Ops framework events.

    For example::

        def test_emitted():
            ctx = Context(
                MyCharm,
                capture_deferred_events=True,
                capture_framework_events=True,
            )
            deferred = ctx.on.update_status().deferred(MyCharm._on_foo)
            ctx.run(ctx.on.start(), State(deferred=[deferred]))

            assert len(ctx.emitted_events) == 5
            assert [e.handle.kind for e in ctx.emitted_events] == [
                'update_status',
                'start',
                'collect_unit_status',
                'pre_commit',
                'commit',
            ]
    """

    requested_storages: dict[str, int]
    """A record of the storages the charm has requested"""

    trace_data: list[ReadableSpan]
    """Trace data generated during the last run.

    Each entry is a :py:class:`opentelemetry.sdk.trace.ReadableSpan`. Tests should not
    rely on the order of this list. Rather, tests could validate parent/child relationships
    or containment by start and end timestamps.
    """

    action_logs: list[str]
    """The logs associated with the action output.

    These are set by the charm with :meth:`ops.ActionEvent.log` and will be empty
    when handling a non-action event.
    """

    action_results: dict[str, Any] | None
    """A key-value mapping assigned by the charm as a result of the action.

    This will be ``None`` if the charm never calls :meth:`ops.ActionEvent.set_results`
    """

    charm_root: str | pathlib.Path | None
    """The charm root directory to use when executing the charm.

    Before running the event, the charm's metadata, config, and action YAML files are
    copied to the charm root. If ``charm_root`` is None, then a temporary directory is
    created and used. To set up the charm root directory with other files, pass in the location of
    the directory to use. If a ``charm_root`` is provided, the tests are also able to inspect the
    content of the directory after the event has run, unlike the temporary directory, which is
    automatically deleted after the run.
    """

    on: CharmEvents
    """The events that this charm can respond to.

    Use this when calling :meth:`run` to specify the event to emit.
    """

    def __init__(
        self,
        charm_type: type[CharmType],
        meta: dict[str, Any] | None = None,
        *,
        actions: dict[str, Any] | None = None,
        config: dict[str, Any] | None = None,
        charm_root: str | pathlib.Path | None = None,
        juju_version: str = _DEFAULT_JUJU_VERSION,
        capture_deferred_events: bool = False,
        capture_framework_events: bool = False,
        app_name: str | None = None,
        unit_id: int | None = 0,
        machine_id: str | None = None,
        availability_zone: str | None = None,
        principal_unit: str | None = None,
        app_trusted: bool = False,
    ):
        """Represents a simulated charm's execution context.

        If the charm, say, expects a ``./src/foo/bar.yaml`` file present relative to the
        execution cwd, you need to use the ``charm_root`` argument. For example::

            import tempfile
            virtual_root = tempfile.TemporaryDirectory()
            local_path = Path(local_path.name)
            (local_path / 'foo').mkdir()
            (local_path / 'foo' / 'bar.yaml').write_text('foo: bar')
            Context(... charm_root=virtual_root).run(...)

        :arg charm_type: the :class:`ops.CharmBase` subclass to handle the event.
        :arg meta: charm metadata to use. Needs to be a valid metadata.yaml format (as a dict).
            If none is provided, we will search for a ``metadata.yaml`` file in the charm root.
        :arg actions: charm actions to use. Needs to be a valid actions.yaml format (as a dict).
            If none is provided, we will search for a ``actions.yaml`` file in the charm root.
        :arg config: charm config to use. Needs to be a valid config.yaml format (as a dict).
            If none is provided, we will search for a ``config.yaml`` file in the charm root.
        :arg juju_version: Juju agent version to simulate.
        :arg app_name: App name that this charm is deployed as. Defaults to the charm name as
            defined in the metadata.
        :arg unit_id: Unit ID that this charm is deployed as, surfaced to the charm in the
            JUJU_UNIT_NAME envvar. Defaults to 0 if unset.
        :arg machine_id: Juju Machine ID that this charm is deployed onto,
            surfaced to the charm as the JUJU_MACHINE_ID envvar.
            Only applicable to machine charms. Unset by default.
        :arg availability_zone: Juju availability zone that this charm is deployed to,
            surfaced to the charm as the JUJU_AVAILABILITY_ZONE envvar.
            Unset by default.
        :arg principal_unit: Juju principal unit name for subordinate charms,
            surfaced to the charm as the JUJU_PRINCIPAL_UNIT envvar.
            Only applicable to subordinate charms. Unset by default.
        :arg app_trusted: whether the charm has Juju trust (deployed with ``--trust`` or added with
            ``juju trust``).
        :arg charm_root: virtual charm filesystem root the charm will be executed with.
        """
        if not any((meta, actions, config)):
            logger.debug('Autoloading charmspec...')
            try:
                spec: _CharmSpec[CharmType] = _CharmSpec.autoload(charm_type)
            except MetadataNotFoundError as e:
                raise ContextSetupError(
                    f'Cannot setup scenario with `charm_type`={charm_type}. '
                    f'Did you forget to pass `meta` to this Context?',
                ) from e

        else:
            if not meta:
                meta = {'name': str(charm_type.__name__)}
            spec = _CharmSpec(
                charm_type=charm_type,
                meta=meta,
                actions=actions,
                config=config,
            )

        self._charm_spec = spec
        self.charm_root = charm_root
        self.juju_version = juju_version
        if juju_version.split('.')[0] == '2':
            logger.warning(
                'Juju 2.x is closed and unsupported. You may encounter inconsistencies.',
            )

        self.app_name = app_name or self._charm_spec.meta.get('name')
        self.unit_id = unit_id
        self._machine_id = machine_id
        self._availability_zone = availability_zone
        self._principal_unit = principal_unit
        self.app_trusted = app_trusted
        self._tmp = tempfile.TemporaryDirectory()

        # config for what events to be captured in emitted_events.
        self.capture_deferred_events = capture_deferred_events
        self.capture_framework_events = capture_framework_events

        # streaming side effects from running an event
        self.juju_log: list[JujuLogLine] = []
        self.app_status_history: list[_EntityStatus] = []
        self.unit_status_history: list[_EntityStatus] = []
        self.exec_history: dict[str, list[ExecArgs]] = {}
        self.workload_version_history: list[str] = []
        self.removed_secret_revisions: list[int] = []
        self.emitted_events: list[ops.EventBase] = []
        self.requested_storages: dict[str, int] = {}
        self.trace_data = []

        # set by Runtime.exec() in self._run()
        self._output_state: State | None = None

        # operations (and embedded tasks) from running actions
        self.action_logs: list[str] = []
        self.action_results: dict[str, Any] | None = None
        self._action_failure_message: str | None = None

        self.on = CharmEvents()

    @property
    @deprecated('Use State.from_context to generate a State from charm metadata.', stacklevel=2)
    def charm_spec(self):
        """Deprecated property for accessing the Context's charm specification.

        The ``_CharmSpec`` class is private and intended for internal use only.
        ``Context.charm_spec`` will be removed in a future major version.

        Consider accessing the charm class or metadata directly, or using
        :meth:`State.from_context` to generate a ``State`` from your charm's metadata.
        """
        return self._charm_spec

    def _set_output_state(self, output_state: State):
        """Hook for Runtime to set the output state."""
        self._output_state = output_state

    def _get_container_root(self, container_name: str):
        """Get the path to a tempdir where this container's simulated root will live."""
        return pathlib.Path(self._tmp.name) / 'containers' / container_name

    def _get_storage_root(self, name: str, index: int) -> pathlib.Path:
        """Get the path to a tempdir where this storage's simulated root will live."""
        storage_root = pathlib.Path(self._tmp.name) / 'storages' / f'{name}-{index}'
        # in the case of _get_container_root, _MockPebbleClient will ensure the dir exists.
        storage_root.mkdir(parents=True, exist_ok=True)
        return storage_root

    def _record_status(self, state: State, is_app: bool):
        """Record the previous status before a status change."""
        if is_app:
            self.app_status_history.append(state.app_status)
        else:
            self.unit_status_history.append(state.unit_status)

    def __call__(self, event: _Event, state: State) -> Manager[CharmType]:
        """Context manager to introspect live charm object before and after the event is emitted.

        Usage::

            ctx = Context(MyCharm)
            with ctx(ctx.on.start(), State()) as manager:
                manager.charm._some_private_setup()
                manager.run()  # this will fire the event
                assert manager.charm._some_private_attribute == "bar"  # noqa

        Args:
            event: the event that the charm will respond to.
            state: the :class:`State` instance to use when handling the event.
        """
        return Manager(self, event, state)

    def run_action(self, action: str, state: State):
        """Use `run()` instead.

        :private:
        """
        raise AttributeError(
            f'call with `ctx.run`, like `ctx.run(ctx.on.action({action!r})` '
            'and find the results in `ctx.action_results`',
        )

    def run(self, event: _Event, state: State) -> State:
        """Trigger a charm execution with an event and a State.

        Calling this function will call ``ops.main`` and set up the context according to the
        specified :class:`State`, then emit the event on the charm.

        :arg event: the event that the charm will respond to. Use the :attr:`on` attribute to
            specify the event; for example: ``ctx.on.start()``.
        :arg state: the :class:`State` instance to use as data source for the hook command
            calls that the charm will invoke when handling the event.
        """
        # Help people transition from Scenario 6:
        if isinstance(event, str):
            event = event.replace('-', '_')  # type: ignore
            if event in (
                'install',
                'start',
                'stop',
                'remove',
                'update_status',
                'config_changed',
                'upgrade_charm',
                'pre_series_upgrade',
                'post_series_upgrade',
                'leader_elected',
                'collect_app_status',
                'collect_unit_status',
            ):  # type: ignore
                suggested = f'{event}()'
            elif event in ('secret_changed', 'secret_rotate'):  # type: ignore
                suggested = f'{event}(my_secret)'
            elif event in ('secret_expired', 'secret_remove'):  # type: ignore
                suggested = f'{event}(my_secret, revision=1)'
            elif event in (
                'relation_created',
                'relation_joined',
                'relation_changed',
                'relation_departed',
                'relation_broken',
            ):  # type: ignore
                suggested = f'{event}(my_relation)'
            elif event in ('storage_attached', 'storage_detaching'):  # type: ignore
                suggested = f'{event}(my_storage)'
            elif event == 'pebble_ready':
                suggested = f'{event}(my_container)'
            elif event == 'pebble_custom_notice':
                suggested = f'{event}(my_container, my_notice)'
            else:
                suggested = 'event()'
            raise TypeError(
                f'call with an event from `ctx.on`, like `ctx.on.{suggested}`',
            )
        if callable(event):
            raise TypeError(
                'You should call the event method. Did you forget to add parentheses?',
            )

        with self._run(event=event, state=state) as ops:
            ops.run()
        # We know that the output state will have been set by this point,
        # so let the type checkers know that too.
        assert self._output_state is not None
        return self._output_state

    @contextmanager
    def _run(self, event: _Event, state: State):
        runtime = Runtime(
            app_name=self.app_name,
            charm_spec=self._charm_spec,
            juju_version=self.juju_version,
            charm_root=self.charm_root,
            unit_id=self.unit_id,
            machine_id=self._machine_id,
            availability_zone=self._availability_zone,
            principal_unit=self._principal_unit,
        )
        if event.action:
            # Reset the logs, failure status, and results, in case the context
            # is reused.
            self.action_logs.clear()
            if self.action_results is not None:
                self.action_results.clear()
            self._action_failure_message = None

        with runtime.exec(
            state=state,
            event=event,
            context=self,
        ) as ops:
            yield ops

        if event.action and self._action_failure_message is not None:
            raise ActionFailed(
                self._action_failure_message,
                state=self._output_state,  # type: ignore
            )



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import dataclasses

import pytest
from scenario._consistency_checker import check_consistency
from scenario.context import Context
from scenario.errors import InconsistentScenarioError
from scenario.state import (
    _RELATION_EVENTS_SUFFIX,
    CheckInfo,
    CloudCredential,
    CloudSpec,
    Container,
    Model,
    Network,
    Notice,
    PeerRelation,
    Relation,
    Resource,
    Secret,
    State,
    Storage,
    StoredState,
    SubordinateRelation,
    _CharmSpec,
    _Event,
)

import ops


class MyCharm(ops.CharmBase):
    pass


def assert_inconsistent(
    state: State,
    event: _Event,
    charm_spec: _CharmSpec,
    juju_version='3.0',
    unit_id=0,
):
    with pytest.raises(InconsistentScenarioError):
        check_consistency(state, event, charm_spec, juju_version, unit_id)


def assert_consistent(
    state: State,
    event: _Event,
    charm_spec: _CharmSpec,
    juju_version='3.0',
    unit_id=0,
):
    check_consistency(state, event, charm_spec, juju_version, unit_id)


def test_base():
    state = State()
    event = _Event('update_status')
    spec = _CharmSpec(MyCharm, {})
    assert_consistent(state, event, spec)


def test_workload_event_without_container():
    assert_inconsistent(
        State(),
        _Event('foo-pebble-ready', container=Container('foo')),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(containers={Container('foo')}),
        _Event('foo-pebble-ready', container=Container('foo')),
        _CharmSpec(MyCharm, {'containers': {'foo': {}}}),
    )
    assert_inconsistent(
        State(),
        _Event('foo-pebble-custom-notice', container=Container('foo')),
        _CharmSpec(MyCharm, {}),
    )
    notice = Notice('example.com/foo')
    assert_consistent(
        State(containers={Container('foo', notices=[notice])}),
        _Event('foo-pebble-custom-notice', container=Container('foo'), notice=notice),
        _CharmSpec(MyCharm, {'containers': {'foo': {}}}),
    )
    assert_inconsistent(
        State(containers={Container('foo')}),
        _Event('foo-pebble-custom-notice', container=Container('foo'), notice=notice),
        _CharmSpec(MyCharm, {'containers': {'foo': {}}}),
    )


def test_check_info_container_matches_event():
    check = CheckInfo(
        'http-check',
        level=ops.pebble.CheckLevel.UNSET,
        startup=ops.pebble.CheckStartup.UNSET,
        threshold=3,
    )
    layer = ops.pebble.Layer({'checks': {'http-check': {'override': 'replace', 'threshold': 3}}})
    for event in ('foo-pebble-check-failed', 'foo-pebble-check-recovered'):
        assert_consistent(
            State(containers={Container('foo', check_infos={check}, layers={'base': layer})}),
            _Event(event, container=Container('foo'), check_info=check),
            _CharmSpec(MyCharm, {'containers': {'foo': {}}}),
        )
        # The check used in the event is missing from the input state.
        assert_inconsistent(
            State(containers={Container('foo', layers={'base': layer})}),
            _Event(event, container=Container('foo'), check_info=check),
            _CharmSpec(MyCharm, {'containers': {'foo': {}}}),
        )
        # The check is in the wrong container.
        assert_inconsistent(
            State(containers={Container('foo', check_infos={check}), Container('bar')}),
            _Event(event, container=Container('bar'), check_info=check),
            _CharmSpec(MyCharm, {'containers': {'foo': {}, 'bar': {}}}),
        )


def test_container_meta_mismatch():
    assert_inconsistent(
        State(containers={Container('bar')}),
        _Event('foo'),
        _CharmSpec(MyCharm, {'containers': {'baz': {}}}),
    )
    assert_consistent(
        State(containers={Container('bar')}),
        _Event('foo'),
        _CharmSpec(MyCharm, {'containers': {'bar': {}}}),
    )


def test_container_in_state_but_no_container_in_meta():
    assert_inconsistent(
        State(containers={Container('bar')}),
        _Event('foo'),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(containers={Container('bar')}),
        _Event('foo'),
        _CharmSpec(MyCharm, {'containers': {'bar': {}}}),
    )


def test_container_not_in_state():
    container = Container('bar')
    assert_inconsistent(
        State(),
        _Event('bar_pebble_ready', container=container),
        _CharmSpec(MyCharm, {'containers': {'bar': {}}}),
    )
    assert_consistent(
        State(containers={container}),
        _Event('bar_pebble_ready', container=container),
        _CharmSpec(MyCharm, {'containers': {'bar': {}}}),
    )


def test_evt_bad_container_name():
    assert_inconsistent(
        State(),
        _Event('foo-pebble-ready', container=Container('bar')),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(containers={Container('bar')}),
        _Event('bar-pebble-ready', container=Container('bar')),
        _CharmSpec(MyCharm, {'containers': {'bar': {}}}),
    )


@pytest.mark.parametrize(
    'check,consistent',
    [
        (
            CheckInfo(
                'chk1',
                level=ops.pebble.CheckLevel.READY,
                startup=ops.pebble.CheckStartup.DISABLED,
                threshold=42,
            ),
            True,
        ),
        (
            CheckInfo(
                'chk1',
                level=ops.pebble.CheckLevel.UNSET,
                startup=ops.pebble.CheckStartup.DISABLED,
                threshold=42,
            ),
            False,
        ),
        (
            CheckInfo(
                'chk1',
                level=ops.pebble.CheckLevel.READY,
                startup=ops.pebble.CheckStartup.ENABLED,
                threshold=42,
            ),
            False,
        ),
        (
            CheckInfo(
                'chk1',
                level=ops.pebble.CheckLevel.READY,
                startup=ops.pebble.CheckStartup.DISABLED,
                threshold=3,
            ),
            False,
        ),
        (CheckInfo('chk2'), False),
    ],
)
def test_checkinfo_matches_layer(check: CheckInfo, consistent: bool):
    layer = ops.pebble.Layer({
        'checks': {
            'chk1': {
                'override': 'replace',
                'level': 'ready',
                'startup': 'disabled',
                'threshold': 42,
                'http': {'url': 'http://localhost:5000/version'},
            }
        }
    })
    state = State(containers={Container('foo', check_infos={check}, layers={'base': layer})})
    asserter = assert_consistent if consistent else assert_inconsistent
    asserter(
        state,
        _Event('foo-pebble-ready', container=Container('foo')),
        _CharmSpec(MyCharm, {'containers': {'foo': {}}}),
    )


@pytest.mark.parametrize('suffix', sorted(_RELATION_EVENTS_SUFFIX))
def test_evt_bad_relation_name(suffix: str):
    assert_inconsistent(
        State(),
        _Event(f'foo{suffix}', relation=Relation('bar')),
        _CharmSpec(MyCharm, {'requires': {'foo': {'interface': 'xxx'}}}),
    )
    relation = Relation('bar')
    assert_consistent(
        State(relations={relation}),
        _Event(f'bar{suffix}', relation=relation),
        _CharmSpec(MyCharm, {'requires': {'bar': {'interface': 'xxx'}}}),
    )


@pytest.mark.parametrize('suffix', sorted(_RELATION_EVENTS_SUFFIX))
def test_evt_no_relation(suffix: str):
    assert_inconsistent(State(), _Event(f'foo{suffix}'), _CharmSpec(MyCharm, {}))
    relation = Relation('bar')
    assert_consistent(
        State(relations={relation}),
        _Event(f'bar{suffix}', relation=relation),
        _CharmSpec(MyCharm, {'requires': {'bar': {'interface': 'xxx'}}}),
    )


def test_config_key_missing_from_meta():
    assert_inconsistent(
        State(config={'foo': True}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(config={'foo': True}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'boolean'}}}),
    )


def test_bad_config_option_type():
    assert_inconsistent(
        State(config={'foo': True}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'string'}}}),
    )
    assert_inconsistent(
        State(config={'foo': True}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {}}}),
    )
    assert_consistent(
        State(config={'foo': True}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'boolean'}}}),
    )


@pytest.mark.parametrize(
    'config_type',
    (
        ('string', 'foo', 1),
        ('int', 1, '1'),
        ('float', 1.0, 1),
        ('boolean', False, 'foo'),
    ),
)
def test_config_types(config_type):
    type_name, valid_value, invalid_value = config_type
    assert_consistent(
        State(config={'foo': valid_value}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': type_name}}}),
    )
    assert_inconsistent(
        State(config={'foo': invalid_value}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': type_name}}}),
    )


@pytest.mark.parametrize('juju_version', ('3.4', '3.5', '4.0'))
def test_config_secret(juju_version):
    assert_consistent(
        State(config={'foo': 'secret:co28kefmp25c77utl3n0'}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'secret'}}}),
        juju_version=juju_version,
    )
    assert_inconsistent(
        State(config={'foo': 1}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'secret'}}}),
    )
    assert_inconsistent(
        State(config={'foo': 'co28kefmp25c77utl3n0'}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'secret'}}}),
    )
    assert_inconsistent(
        State(config={'foo': 'secret:secret'}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'secret'}}}),
    )
    assert_inconsistent(
        State(config={'foo': 'secret:co28kefmp25c77utl3n!'}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'secret'}}}),
    )


@pytest.mark.parametrize('juju_version', ('2.9', '3.3'))
def test_config_secret_old_juju(juju_version):
    assert_inconsistent(
        State(config={'foo': 'secret:co28kefmp25c77utl3n0'}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}, config={'options': {'foo': {'type': 'secret'}}}),
        juju_version=juju_version,
    )


@pytest.mark.skip(
    "The right exception is raised but pytest.raises doesn't catch it - figure this out!"
)
@pytest.mark.parametrize('bad_v', ('1.0', '0', '1.2', '2.35.42', '2.99.99', '2.99'))
def test_secrets_jujuv_bad(bad_v):
    secret = Secret({'a': 'b'})
    assert_inconsistent(
        State(secrets={secret}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}),
        bad_v,
    )
    assert_inconsistent(
        State(secrets={secret}),
        _Event('secret_changed', secret=secret),
        _CharmSpec(MyCharm, {}),
        bad_v,
    )

    assert_inconsistent(
        State(),
        _Event('secret_changed', secret=secret),
        _CharmSpec(MyCharm, {}),
        bad_v,
    )


@pytest.mark.parametrize('good_v', ('3.0', '3.1', '3', '3.33', '4', '100'))
def test_secrets_jujuv_good(good_v):
    assert_consistent(
        State(secrets={Secret({'a': 'b'})}),
        _Event('bar'),
        _CharmSpec(MyCharm, {}),
        good_v,
    )


def test_secret_not_in_state():
    secret = Secret({'a': 'b'})
    assert_inconsistent(
        State(),
        _Event('secret_changed', secret=secret),
        _CharmSpec(MyCharm, {}),
    )
    assert_consistent(
        State(secrets={secret}),
        _Event('secret_changed', secret=secret),
        _CharmSpec(MyCharm, {}),
    )


def test_peer_relation_consistency():
    assert_inconsistent(
        State(relations={Relation('foo')}),
        _Event('bar'),
        _CharmSpec(MyCharm, {'peers': {'foo': {'interface': 'bar'}}}),
    )
    assert_consistent(
        State(relations={PeerRelation('foo')}),
        _Event('bar'),
        _CharmSpec(MyCharm, {'peers': {'foo': {'interface': 'bar'}}}),
    )


def test_duplicate_endpoints_inconsistent():
    assert_inconsistent(
        State(),
        _Event('bar'),
        _CharmSpec(
            MyCharm,
            {
                'requires': {'foo': {'interface': 'bar'}},
                'provides': {'foo': {'interface': 'baz'}},
            },
        ),
    )


def test_sub_relation_consistency():
    assert_inconsistent(
        State(relations={Relation('foo')}),
        _Event('bar'),
        _CharmSpec(
            MyCharm,
            {'requires': {'foo': {'interface': 'bar', 'scope': 'container'}}},
        ),
    )

    assert_consistent(
        State(relations={SubordinateRelation('foo')}),
        _Event('bar'),
        _CharmSpec(
            MyCharm,
            {'requires': {'foo': {'interface': 'bar', 'scope': 'container'}}},
        ),
    )


def test_relation_sub_inconsistent():
    assert_inconsistent(
        State(relations={SubordinateRelation('foo')}),
        _Event('bar'),
        _CharmSpec(MyCharm, {'requires': {'foo': {'interface': 'bar'}}}),
    )


def test_relation_not_in_state():
    relation = Relation('foo')
    assert_inconsistent(
        State(),
        _Event('foo_relation_changed', relation=relation),
        _CharmSpec(MyCharm, {'requires': {'foo': {'interface': 'bar'}}}),
    )
    assert_consistent(
        State(relations={relation}),
        _Event('foo_relation_changed', relation=relation),
        _CharmSpec(MyCharm, {'requires': {'foo': {'interface': 'bar'}}}),
    )


def test_action_not_in_meta_inconsistent():
    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'foo': {}})
    assert_inconsistent(
        State(),
        ctx.on.action('foo', params={'bar': 'baz'}),
        _CharmSpec(MyCharm, meta={}, actions={}),
    )


def test_action_meta_type_inconsistent():
    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'foo': {}})
    assert_inconsistent(
        State(),
        ctx.on.action('foo', params={'bar': 'baz'}),
        _CharmSpec(MyCharm, meta={}, actions={'foo': {'params': {'bar': {'type': 'zabazaba'}}}}),
    )

    assert_inconsistent(
        State(),
        ctx.on.action('foo', params={'bar': 'baz'}),
        _CharmSpec(MyCharm, meta={}, actions={'foo': {'params': {'bar': {}}}}),
    )

    assert_inconsistent(
        State(),
        ctx.on.action('foo', params={'bar': 'baz'}),
        _CharmSpec(MyCharm, meta={}, actions={'foo': None}),
    )


def test_action_name():
    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'foo': {}})

    assert_consistent(
        State(),
        ctx.on.action('foo', params={'bar': 'baz'}),
        _CharmSpec(MyCharm, meta={}, actions={'foo': {'params': {'bar': {'type': 'string'}}}}),
    )

    assert_consistent(
        State(),
        ctx.on.action('foo', params={}),
        _CharmSpec(MyCharm, meta={}, actions={'foo': None}),
    )

    assert_inconsistent(
        State(),
        _Event('box_action', action=ctx.on.action('foo', params={'bar': 'baz'})),
        _CharmSpec(MyCharm, meta={}, actions={'foo': {}}),
    )


_ACTION_TYPE_CHECKS = [
    ('string', 'baz', None),
    ('boolean', True, 'baz'),
    ('integer', 42, 1.5),
    ('number', 28.8, 'baz'),
    ('array', ['a', 'b', 'c'], 1.5),  # A string is an acceptable array.
    ('object', {'k': 'v'}, 'baz'),
]


@pytest.mark.parametrize('ptype,good,bad', _ACTION_TYPE_CHECKS)
def test_action_params_type(ptype, good, bad):
    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'foo': {}})
    assert_consistent(
        State(),
        ctx.on.action('foo', params={'bar': good}),
        _CharmSpec(MyCharm, meta={}, actions={'foo': {'params': {'bar': {'type': ptype}}}}),
    )
    if bad is not None:
        assert_inconsistent(
            State(),
            ctx.on.action('foo', params={'bar': bad}),
            _CharmSpec(MyCharm, meta={}, actions={'foo': {'params': {'bar': {'type': ptype}}}}),
        )


def test_duplicate_relation_ids():
    assert_inconsistent(
        State(relations={Relation('foo', id=1), Relation('bar', id=1)}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'requires': {'foo': {'interface': 'foo'}, 'bar': {'interface': 'bar'}}},
        ),
    )


def test_relation_without_endpoint():
    assert_inconsistent(
        State(relations={Relation('foo', id=1), Relation('bar', id=1)}),
        _Event('start'),
        _CharmSpec(MyCharm, meta={'name': 'charlemagne'}),
    )

    assert_consistent(
        State(relations={Relation('foo', id=1), Relation('bar', id=2)}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'requires': {'foo': {'interface': 'foo'}, 'bar': {'interface': 'bar'}}},
        ),
    )


def test_storage_event():
    storage = Storage('foo')
    assert_inconsistent(
        State(storages={storage}),
        _Event('foo-storage-attached'),
        _CharmSpec(MyCharm, meta={'name': 'rupert'}),
    )
    assert_inconsistent(
        State(storages={storage}),
        _Event('foo-storage-attached'),
        _CharmSpec(MyCharm, meta={'name': 'rupert', 'storage': {'foo': {'type': 'filesystem'}}}),
    )


def test_storage_states():
    storage1 = Storage('foo', index=1)
    storage2 = Storage('foo', index=1)

    assert_inconsistent(
        State(storages={storage1, storage2}),
        _Event('start'),
        _CharmSpec(MyCharm, meta={'name': 'everett'}),
    )
    assert_consistent(
        State(storages={storage1, dataclasses.replace(storage2, index=2)}),
        _Event('start'),
        _CharmSpec(MyCharm, meta={'name': 'frank', 'storage': {'foo': {'type': 'filesystem'}}}),
    )
    assert_consistent(
        State(storages={storage1, dataclasses.replace(storage2, name='marx')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={
                'name': 'engels',
                'storage': {
                    'foo': {'type': 'filesystem'},
                    'marx': {'type': 'filesystem'},
                },
            },
        ),
    )


def test_storage_not_in_state():
    storage = Storage('foo')
    assert_inconsistent(
        State(),
        _Event('foo_storage_attached', storage=storage),
        _CharmSpec(
            MyCharm,
            meta={'name': 'sam', 'storage': {'foo': {'type': 'filesystem'}}},
        ),
    )
    assert_consistent(
        State(storages=[storage]),
        _Event('foo_storage_attached', storage=storage),
        _CharmSpec(
            MyCharm,
            meta={'name': 'sam', 'storage': {'foo': {'type': 'filesystem'}}},
        ),
    )


def test_resource_states():
    # happy path
    assert_consistent(
        State(resources={Resource(name='foo', path='/foo/bar.yaml')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'yamlman', 'resources': {'foo': {'type': 'oci-image'}}},
        ),
    )

    # no resources in state but some in meta: OK. Not realistic wrt juju but fine for testing
    assert_consistent(
        State(),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'yamlman', 'resources': {'foo': {'type': 'oci-image'}}},
        ),
    )

    # resource not defined in meta
    assert_inconsistent(
        State(resources={Resource(name='bar', path='/foo/bar.yaml')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'yamlman', 'resources': {'foo': {'type': 'oci-image'}}},
        ),
    )

    assert_inconsistent(
        State(resources={Resource(name='bar', path='/foo/bar.yaml')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'yamlman'},
        ),
    )


def test_networks_consistency():
    assert_inconsistent(
        State(networks={Network('foo')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'wonky'},
        ),
    )

    assert_inconsistent(
        State(networks={Network('foo')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={
                'name': 'pinky',
                'extra-bindings': {'foo': {}},
                'requires': {'foo': {'interface': 'bar'}},
            },
        ),
    )

    assert_consistent(
        State(networks={Network('foo')}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={
                'name': 'pinky',
                'extra-bindings': {'foo': {}},
                'requires': {'bar': {'interface': 'bar'}},
            },
        ),
    )


def test_cloudspec_consistency():
    cloud_spec = CloudSpec(
        name='localhost',
        type='lxd',
        endpoint='https://127.0.0.1:8443',
        credential=CloudCredential(
            auth_type='clientcertificate',
            attributes={
                'client-cert': 'foo',
                'client-key': 'bar',
                'server-cert': 'baz',
            },
        ),
    )

    assert_consistent(
        State(model=Model(name='lxd-model', type='lxd', cloud_spec=cloud_spec)),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'MyVMCharm'},
        ),
    )

    assert_inconsistent(
        State(model=Model(name='k8s-model', type='kubernetes', cloud_spec=cloud_spec)),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={'name': 'MyK8sCharm'},
        ),
    )


def test_storedstate_consistency():
    assert_consistent(
        State(
            stored_states={
                StoredState(content={'foo': 'bar'}),
                StoredState(name='my_stored_state', content={'foo': 1}),
                StoredState(owner_path='MyCharmLib', content={'foo': None}),
                StoredState(owner_path='OtherCharmLib', content={'foo': (1, 2, 3)}),
            }
        ),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={
                'name': 'foo',
            },
        ),
    )
    assert_inconsistent(
        State(stored_states={StoredState(owner_path=None, content={'secret': Secret({})})}),
        _Event('start'),
        _CharmSpec(
            MyCharm,
            meta={
                'name': 'foo',
            },
        ),
    )


@pytest.mark.parametrize('unit_id', (0, 1, 42))
def test_own_unit_in_peers_data_consistency(unit_id: int):
    assert_consistent(
        State(relations={PeerRelation('foo', peers_data={unit_id + 1: {'foo': 'bar'}})}),
        _Event('start'),
        _CharmSpec(MyCharm, {'peers': {'foo': {'interface': 'bar'}}}),
        unit_id=unit_id,
    )
    assert_inconsistent(
        State(
            relations={
                PeerRelation(
                    'foo', peers_data={unit_id: {'foo': 'bar'}, unit_id + 1: {'foo': 'baz'}}
                )
            }
        ),
        _Event('start'),
        _CharmSpec(MyCharm, {'peers': {'foo': {'interface': 'bar'}}}),
        unit_id=unit_id,
    )



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import copy
import datetime
import typing

import pytest
import scenario

import ops

META = {
    'name': 'context-charm',
    'containers': {
        'bar': {},
    },
    'requires': {
        'baz': {
            'interface': 'charmlink',
        }
    },
    'storage': {
        'foo': {
            'type': 'filesystem',
        }
    },
}
ACTIONS = {
    'act': {
        'params': {
            'param': {
                'description': 'some parameter',
                'type': 'string',
                'default': '',
            }
        }
    },
}


class ContextCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        self.observed: list[ops.EventBase] = []
        for event in self.on.events().values():
            framework.observe(event, self._on_event)

    def _on_event(self, event: ops.EventBase):
        self.observed.append(event)


@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('install', ops.InstallEvent),
        ('start', ops.StartEvent),
        ('stop', ops.StopEvent),
        ('remove', ops.RemoveEvent),
        ('update_status', ops.UpdateStatusEvent),
        ('config_changed', ops.ConfigChangedEvent),
        ('upgrade_charm', ops.UpgradeCharmEvent),
        ('leader_elected', ops.LeaderElectedEvent),
    ],
)
def test_simple_events(event_name: str, event_kind: type[ops.EventBase]):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.install(), state)
    with ctx(getattr(ctx.on, event_name)(), scenario.State()) as mgr:
        mgr.run()
        juju_event, status = mgr.charm.observed
        assert isinstance(juju_event, event_kind)
        assert isinstance(status, ops.CollectStatusEvent)


@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('pre_series_upgrade', ops.PreSeriesUpgradeEvent),
        ('post_series_upgrade', ops.PostSeriesUpgradeEvent),
    ],
)
def test_simple_deprecated_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.pre_series_upgrade(), state)
    with pytest.warns(DeprecationWarning):
        with ctx(getattr(ctx.on, event_name)(), scenario.State()) as mgr:
            mgr.run()
            deprecated_event, collect_status = mgr.charm.observed
            assert isinstance(deprecated_event, event_kind)
            assert isinstance(collect_status, ops.CollectStatusEvent)


@pytest.mark.parametrize('as_kwarg', [True, False])
@pytest.mark.parametrize(
    'event_name,event_kind,owner',
    [
        ('secret_changed', ops.SecretChangedEvent, None),
        ('secret_rotate', ops.SecretRotateEvent, 'app'),
    ],
)
def test_simple_secret_events(as_kwarg, event_name, event_kind, owner):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    secret = scenario.Secret({'password': 'xxxx'}, owner=owner)
    state_in = scenario.State(secrets={secret})
    # These look like:
    #   ctx.run(ctx.on.secret_changed(secret=secret), state)
    # The secret must always be passed because the same event name is used for
    # all secrets.
    if as_kwarg:
        args = ()
        kwargs = {'secret': secret}
    else:
        args = (secret,)
        kwargs = {}
    with ctx(getattr(ctx.on, event_name)(*args, **kwargs), state_in) as mgr:
        mgr.run()
        secret_event, collect_status = mgr.charm.observed
        assert isinstance(secret_event, event_kind)
        assert secret_event.secret.id == secret.id
        assert isinstance(collect_status, ops.CollectStatusEvent)


@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('secret_expired', ops.SecretExpiredEvent),
        ('secret_remove', ops.SecretRemoveEvent),
    ],
)
def test_revision_secret_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    secret = scenario.Secret(
        tracked_content={'password': 'yyyy'},
        latest_content={'password': 'xxxx'},
        owner='app',
    )
    state_in = scenario.State(secrets={secret})
    # These look like:
    #   ctx.run(ctx.on.secret_expired(secret=secret, revision=revision), state)
    # The secret and revision must always be passed because the same event name
    # is used for all secrets.
    with ctx(getattr(ctx.on, event_name)(secret, revision=42), state_in) as mgr:
        mgr.run()
        secret_event, collect_status = mgr.charm.observed
        assert isinstance(secret_event, event_kind)
        assert secret_event.secret.id == secret.id
        assert secret_event.revision == 42
        assert isinstance(collect_status, ops.CollectStatusEvent)


@pytest.mark.parametrize('event_name', ['secret_expired', 'secret_remove'])
def test_revision_secret_events_as_positional_arg(event_name):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    secret = scenario.Secret(
        tracked_content={'password': 'yyyy'},
        latest_content={'password': 'xxxx'},
        owner=None,
    )
    state_in = scenario.State(secrets={secret})
    with pytest.raises(TypeError):
        ctx.run(getattr(ctx.on, event_name)(secret, 42), state_in)


@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('storage_attached', ops.StorageAttachedEvent),
        ('storage_detaching', ops.StorageDetachingEvent),
    ],
)
def test_storage_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    storage = scenario.Storage('foo')
    state_in = scenario.State(storages=[storage])
    # These look like:
    #   ctx.run(ctx.on.storage_attached(storage), state)
    with ctx(getattr(ctx.on, event_name)(storage), state_in) as mgr:
        mgr.run()
        storage_event, collect_status = mgr.charm.observed
        assert isinstance(storage_event, event_kind)
        assert storage_event.storage.name == storage.name
        assert storage_event.storage.index == storage.index
        assert isinstance(collect_status, ops.CollectStatusEvent)


def test_action_event_no_params():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.action(action_name), state)
    with ctx(ctx.on.action('act'), scenario.State()) as mgr:
        mgr.run()
        action_event, collect_status = mgr.charm.observed
        assert isinstance(action_event, ops.ActionEvent)
        assert isinstance(collect_status, ops.CollectStatusEvent)


def test_action_event_with_params():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    # These look like:
    #   ctx.run(ctx.on.action(action=action), state)
    # So that any parameters can be included and the ID can be customised.
    call_event = ctx.on.action('act', params={'param': 'hello'})
    with ctx(call_event, scenario.State()) as mgr:
        mgr.run()
        action_event, collect_status = mgr.charm.observed
        assert isinstance(action_event, ops.ActionEvent)
        assert action_event.id == call_event.action.id
        assert action_event.params['param'] == call_event.action.params['param']
        assert isinstance(collect_status, ops.CollectStatusEvent)


def test_pebble_ready_event():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    container = scenario.Container('bar', can_connect=True)
    state_in = scenario.State(containers=[container])
    # These look like:
    #   ctx.run(ctx.on.pebble_ready(container), state)
    with ctx(ctx.on.pebble_ready(container), state_in) as mgr:
        mgr.run()
        pebble_ready_event, collect_status = mgr.charm.observed
        assert isinstance(pebble_ready_event, ops.PebbleReadyEvent)
        assert pebble_ready_event.workload.name == container.name
        assert isinstance(collect_status, ops.CollectStatusEvent)


@pytest.mark.parametrize('as_kwarg', [True, False])
@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('relation_created', ops.RelationCreatedEvent),
        ('relation_broken', ops.RelationBrokenEvent),
    ],
)
def test_relation_app_events(as_kwarg, event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation('baz')
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.relation_created(relation), state)
    if as_kwarg:
        args = ()
        kwargs = {'relation': relation}
    else:
        args = (relation,)
        kwargs = {}
    with ctx(getattr(ctx.on, event_name)(*args, **kwargs), state_in) as mgr:
        mgr.run()
        relation_event, collect_status = mgr.charm.observed
        assert isinstance(relation_event, event_kind)
        assert relation_event.relation.id == relation.id
        assert relation_event.app.name == relation.remote_app_name
        assert relation_event.unit is None
        assert isinstance(collect_status, ops.CollectStatusEvent)


def test_relation_complex_name():
    meta = copy.deepcopy(META)
    meta['requires']['foo-bar-baz'] = {'interface': 'another-one'}
    ctx = scenario.Context(ContextCharm, meta=meta, actions=ACTIONS)
    relation = scenario.Relation('foo-bar-baz')
    state_in = scenario.State(relations=[relation])
    with ctx(ctx.on.relation_created(relation), state_in) as mgr:
        mgr.run()
        relation_event, collect_status = mgr.charm.observed
        assert isinstance(relation_event, ops.RelationCreatedEvent)
        assert relation_event.relation.id == relation.id
        assert relation_event.app.name == relation.remote_app_name
        assert relation_event.unit is None
        assert isinstance(collect_status, ops.CollectStatusEvent)


@pytest.mark.parametrize('event_name', ['relation_created', 'relation_broken'])
def test_relation_events_as_positional_arg(event_name):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation('baz')
    state_in = scenario.State(relations=[relation])
    with pytest.raises(TypeError):
        ctx.run(getattr(ctx.on, event_name)(relation, 0), state_in)


@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('relation_joined', ops.RelationJoinedEvent),
        ('relation_changed', ops.RelationChangedEvent),
    ],
)
def test_relation_unit_events_default_unit(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation('baz', remote_units_data={1: {'x': 'y'}})
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.baz_relation_changed, state)
    # The unit is chosen automatically.
    with ctx(getattr(ctx.on, event_name)(relation), state_in) as mgr:
        mgr.run()
        relation_event, collect_status = mgr.charm.observed
        assert isinstance(relation_event, event_kind)
        assert relation_event.relation.id == relation.id
        assert relation_event.app.name == relation.remote_app_name
        assert relation_event.unit.name == 'remote/1'
        assert isinstance(collect_status, ops.CollectStatusEvent)


@pytest.mark.parametrize(
    'event_name, event_kind',
    [
        ('relation_joined', ops.RelationJoinedEvent),
        ('relation_changed', ops.RelationChangedEvent),
    ],
)
def test_relation_unit_events(event_name, event_kind):
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation('baz', remote_units_data={1: {'x': 'y'}, 2: {'x': 'z'}})
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.baz_relation_changed(unit=unit_ordinal), state)
    with ctx(getattr(ctx.on, event_name)(relation, remote_unit=2), state_in) as mgr:
        mgr.run()
        relation_event, collect_status = mgr.charm.observed
        assert isinstance(relation_event, event_kind)
        assert relation_event.relation.id == relation.id
        assert relation_event.app.name == relation.remote_app_name
        assert relation_event.unit.name == 'remote/2'
        assert isinstance(collect_status, ops.CollectStatusEvent)


def test_relation_departed_event():
    ctx = scenario.Context(ContextCharm, meta=META, actions=ACTIONS)
    relation = scenario.Relation('baz')
    state_in = scenario.State(relations=[relation])
    # These look like:
    #   ctx.run(ctx.on.baz_relation_departed(unit=unit_num, departing_unit=unit_num), state)
    with ctx(ctx.on.relation_departed(relation, remote_unit=2, departing_unit=1), state_in) as mgr:
        mgr.run()
        relation_event, collect_status = mgr.charm.observed
        assert isinstance(relation_event, ops.RelationDepartedEvent)
        assert relation_event.relation.id == relation.id
        assert relation_event.app.name == relation.remote_app_name
        assert relation_event.unit.name == 'remote/2'
        assert relation_event.departing_unit.name == 'remote/1'
        assert isinstance(collect_status, ops.CollectStatusEvent)


class CustomEvent(ops.EventBase):
    pass


class CustomEventWithArgs(CustomEvent):
    arg0: str
    arg1: int

    def __init__(self, handle: ops.Handle, arg0: str = '', arg1: int = 0):
        super().__init__(handle)
        self.arg0 = arg0
        self.arg1 = arg1

    def snapshot(self):
        base = super().snapshot()
        base.update({'arg0': self.arg0, 'arg1': self.arg1})
        return base

    def restore(self, snapshot: dict[str, typing.Any]):
        super().restore(snapshot)
        self.arg0 = snapshot['arg0']
        self.arg1 = snapshot['arg1']


class CustomRelationEvent(ops.RelationChangedEvent):
    pass


class CustomEventWithScenarioArgs(CustomEvent):
    cloudcredential: ops.CloudCredential
    cloudspec: ops.CloudSpec
    secret: ops.Secret
    relation: ops.Relation
    peerrelation: ops.Relation
    subordinaterelation: ops.Relation
    notice: ops.pebble.Notice
    checkinfo: ops.pebble.CheckInfo
    container: ops.Container
    errorstatus: ops.ErrorStatus
    activestatus: ops.ActiveStatus
    blockedstatus: ops.BlockedStatus
    maintenancestatus: ops.MaintenanceStatus
    waitingstatus: ops.WaitingStatus
    tcpport: ops.Port
    udpport: ops.Port
    icmpport: ops.Port
    storage: ops.Storage

    def __init__(self, handle: ops.Handle, **kwargs: typing.Any):
        super().__init__(handle)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def snapshot(self):
        base = super().snapshot()
        # This loses a lot of the details, but for the test all we care about is
        # that the type and the 'primary key' are correct.
        base['cloudcredential'] = self.cloudcredential.auth_type
        base['cloudspec'] = self.cloudspec.name
        base['secret'] = self.secret.id
        base['relation'] = self.relation.name
        base['peerrelation'] = self.peerrelation.name
        base['subordinaterelation'] = self.subordinaterelation.name
        base['relation_id'] = self.relation.id
        base['peerrelation_id'] = self.peerrelation.id
        base['subordinaterelation_id'] = self.subordinaterelation.id
        base['notice_id'] = self.notice.id
        base['notice_key'] = self.notice.key
        base['checkinfo'] = self.checkinfo.name
        base['container'] = self.container.name
        base['errorstatus'] = self.errorstatus.message
        base['activestatus'] = self.activestatus.message
        base['blockedstatus'] = self.blockedstatus.message
        base['maintenancestatus'] = self.maintenancestatus.message
        base['waitingstatus'] = self.waitingstatus.message
        base['tcpport'] = self.tcpport.port
        base['udpport'] = self.udpport.port
        base['storage_name'] = self.storage.name
        base['storage_index'] = self.storage.index
        return base

    def restore(self, snapshot: dict[str, typing.Any]):
        super().restore(snapshot)
        self.cloudcredential = ops.CloudCredential(auth_type=snapshot['cloudcredential'])
        self.cloudspec = ops.CloudSpec('', snapshot['cloudspec'])
        self.secret = self.framework.model.get_secret(id=snapshot['secret'])
        relation = self.framework.model.get_relation(
            snapshot['relation'], relation_id=snapshot['relation_id']
        )
        assert relation is not None
        self.relation = relation
        peerrelation = self.framework.model.get_relation(
            snapshot['peerrelation'], relation_id=snapshot['peerrelation_id']
        )
        assert peerrelation is not None
        self.peerrelation = peerrelation
        subordinaterelation = self.framework.model.get_relation(
            snapshot['subordinaterelation'],
            relation_id=snapshot['subordinaterelation_id'],
        )
        assert subordinaterelation is not None
        self.subordinaterelation = subordinaterelation
        now = datetime.datetime.now()
        self.notice = ops.pebble.Notice(
            snapshot['notice_id'],
            None,
            '',
            snapshot['notice_key'],
            now,
            now,
            now,
            1,
        )
        self.checkinfo = ops.pebble.CheckInfo(snapshot['checkinfo'], None, '')
        self.container = self.framework.model.unit.get_container(snapshot['container'])
        self.errorstatus = ops.ErrorStatus(message=snapshot['errorstatus'])
        self.activestatus = ops.ActiveStatus(message=snapshot['activestatus'])
        self.blockedstatus = ops.BlockedStatus(message=snapshot['blockedstatus'])
        self.maintenancestatus = ops.MaintenanceStatus(message=snapshot['maintenancestatus'])
        self.waitingstatus = ops.WaitingStatus(message=snapshot['waitingstatus'])
        self.tcpport = ops.Port(protocol='tcp', port=snapshot['tcpport'])
        self.udpport = ops.Port(protocol='udp', port=snapshot['udpport'])
        self.icmpport = ops.Port(protocol='icmp', port=None)
        for storage in self.framework.model.storages[snapshot['storage_name']]:
            if storage.index == snapshot['storage_index']:
                self.storage = storage
                break


class CustomEvents(ops.ObjectEvents):
    foo_started = ops.EventSource(CustomEvent)
    foo_changed = ops.EventSource(CustomEventWithArgs)
    foo_relation_changed = ops.EventSource(CustomRelationEvent)
    state_event = ops.EventSource(CustomEventWithScenarioArgs)


class MyConsumer(ops.Object):
    on = CustomEvents()  # type: ignore

    def __init__(self, charm: ops.CharmBase):
        super().__init__(charm, 'my-consumer')


class CustomCharm(ContextCharm):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        self.consumer = MyConsumer(self)
        framework.observe(self.consumer.on.foo_started, self._on_event)
        framework.observe(self.consumer.on.foo_changed, self._on_event)
        framework.observe(self.consumer.on.foo_relation_changed, self._on_event)
        framework.observe(self.consumer.on.state_event, self._on_event)


def test_custom_event_no_args():
    ctx = scenario.Context(CustomCharm, meta=META, actions=ACTIONS)
    with ctx(ctx.on.custom(MyConsumer.on.foo_started), scenario.State()) as mgr:
        mgr.run()
        custom_event, collect_status = mgr.charm.observed
        assert isinstance(collect_status, ops.CollectStatusEvent)
        assert isinstance(custom_event, CustomEvent)


def test_custom_event_with_args():
    ctx = scenario.Context(CustomCharm, meta=META, actions=ACTIONS)
    with ctx(
        ctx.on.custom(MyConsumer.on.foo_changed, 'foo', arg1=42),
        scenario.State(),
    ) as mgr:
        mgr.run()
        custom_event, collect_status = mgr.charm.observed
        assert isinstance(collect_status, ops.CollectStatusEvent)
        assert isinstance(custom_event, CustomEventWithArgs)
        assert custom_event.arg0 == 'foo'
        assert custom_event.arg1 == 42


def test_custom_event_is_hookevent():
    ctx = scenario.Context(CustomCharm, meta=META, actions=ACTIONS)
    with pytest.raises(ValueError):
        ctx.on.custom(MyConsumer.on.foo_relation_changed)


def test_custom_event_with_scenario_args():
    meta = META.copy()
    meta['requires']['endpoint'] = {'interface': 'int1'}
    meta['requires']['sub-endpoint'] = {'interface': 'int2', 'scope': 'container'}
    meta['peers'] = {'peer-endpoint': {'interface': 'int3'}}
    meta['containers']['container'] = {}
    meta['storage']['store'] = {'type': 'filesystem'}
    ctx = scenario.Context(CustomCharm, meta=meta, actions=ACTIONS)

    cloudcredential = scenario.CloudCredential(auth_type='auth')
    cloudspec = scenario.CloudSpec('cloud')
    secret = scenario.Secret({'password': 'xxxx'})
    relation = scenario.Relation('endpoint')
    peerrelation = scenario.PeerRelation('peer-endpoint')
    subordinaterelation = scenario.SubordinateRelation('sub-endpoint')
    notice = scenario.Notice('key.example.com')
    layer = ops.pebble.Layer({
        'checks': {'check1': {'override': 'replace', 'startup': 'enabled', 'threshold': 3}}
    })
    checkinfo = scenario.CheckInfo('check1', level=ops.pebble.CheckLevel.UNSET)
    container = scenario.Container(
        'container', notices=[notice], check_infos={checkinfo}, layers={'layer': layer}
    )
    errorstatus = scenario.ErrorStatus('error')
    activestatus = scenario.ActiveStatus('working')
    blockedstatus = scenario.BlockedStatus('blocked')
    maintenancestatus = scenario.MaintenanceStatus('maintaining')
    waitingstatus = scenario.WaitingStatus('waiting')
    tcpport = scenario.TCPPort(8000)
    udpport = scenario.UDPPort(8001)
    icmpport = scenario.ICMPPort()
    storage = scenario.Storage('store')

    state = scenario.State(
        secrets={secret},
        relations={relation, peerrelation, subordinaterelation},
        containers={container},
        storages={storage},
    )

    with ctx(
        ctx.on.custom(
            MyConsumer.on.state_event,
            cloudcredential=cloudcredential,
            cloudspec=cloudspec,
            secret=secret,
            relation=relation,
            peerrelation=peerrelation,
            subordinaterelation=subordinaterelation,
            notice=notice,
            checkinfo=checkinfo,
            container=container,
            errorstatus=errorstatus,
            activestatus=activestatus,
            blockedstatus=blockedstatus,
            maintenancestatus=maintenancestatus,
            waitingstatus=waitingstatus,
            tcpport=tcpport,
            udpport=udpport,
            icmpport=icmpport,
            storage=storage,
        ),
        state,
    ) as mgr:
        mgr.run()
        evt, cs = mgr.charm.observed
        assert isinstance(cs, ops.CollectStatusEvent)
        assert isinstance(evt, CustomEventWithScenarioArgs)
        assert isinstance(evt.cloudcredential, ops.CloudCredential)
        assert evt.cloudcredential.auth_type == cloudcredential.auth_type
        assert isinstance(evt.cloudspec, ops.CloudSpec)
        assert evt.cloudspec.name == cloudspec.name
        assert isinstance(evt.secret, ops.Secret)
        assert evt.secret.id == secret.id
        assert isinstance(evt.relation, ops.Relation)
        assert evt.relation.name == relation.endpoint
        assert isinstance(evt.peerrelation, ops.Relation)
        assert evt.peerrelation.name == peerrelation.endpoint
        assert isinstance(evt.subordinaterelation, ops.Relation)
        assert evt.subordinaterelation.name == subordinaterelation.endpoint
        assert isinstance(evt.notice, ops.pebble.Notice)
        assert evt.notice.key == notice.key
        assert isinstance(evt.checkinfo, ops.pebble.CheckInfo)
        assert evt.checkinfo.name == checkinfo.name
        assert isinstance(evt.container, ops.Container)
        assert evt.container.name == container.name
        assert isinstance(evt.errorstatus, ops.ErrorStatus)
        assert evt.errorstatus.message == errorstatus.message
        assert isinstance(evt.activestatus, ops.ActiveStatus)
        assert evt.activestatus.message == activestatus.message
        assert isinstance(evt.blockedstatus, ops.BlockedStatus)
        assert evt.blockedstatus.message == blockedstatus.message
        assert isinstance(evt.maintenancestatus, ops.MaintenanceStatus)
        assert evt.maintenancestatus.message == maintenancestatus.message
        assert isinstance(evt.waitingstatus, ops.WaitingStatus)
        assert evt.waitingstatus.message == waitingstatus.message
        assert isinstance(evt.tcpport, ops.Port)
        assert evt.tcpport.protocol == 'tcp'
        assert evt.tcpport.port == tcpport.port
        assert isinstance(evt.udpport, ops.Port)
        assert evt.udpport.protocol == 'udp'
        assert evt.udpport.port == udpport.port
        assert isinstance(evt.icmpport, ops.Port)
        assert evt.icmpport.protocol == 'icmp'
        assert isinstance(evt.storage, ops.Storage)
        assert evt.storage.name == storage.name


class OtherEvent(ops.EventBase):
    pass


class OtherEvents(ops.ObjectEvents):
    foo_changed = ops.EventSource(OtherEvent)


class OtherConsumer(ops.Object):
    on = OtherEvents()  # type: ignore

    def __init__(self, charm: ops.CharmBase, relation_name: str):
        super().__init__(charm, relation_name)


class TwoLibraryCharm(ContextCharm):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        self.consumer1 = MyConsumer(self)
        self.consumer2 = OtherConsumer(self, 'some-relation')
        framework.observe(self.consumer1.on.foo_changed, self._on_event)
        framework.observe(self.consumer2.on.foo_changed, self._on_event)


def test_custom_event_two_libraries():
    ctx = scenario.Context(TwoLibraryCharm, meta=META, actions=ACTIONS)

    with ctx(ctx.on.custom(MyConsumer.on.foo_changed), scenario.State()) as mgr:
        mgr.run()
        evt, cs = mgr.charm.observed
        assert isinstance(cs, ops.CollectStatusEvent)
        assert isinstance(evt, CustomEvent)

    with ctx(ctx.on.custom(OtherConsumer.on.foo_changed), scenario.State()) as mgr:
        mgr.run()
        evt, collect_status = mgr.charm.observed
        assert isinstance(collect_status, ops.CollectStatusEvent)
        assert isinstance(evt, OtherEvent)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import dataclasses

import pytest
from scenario.state import BlockedStatus as ScenarioBlockedStatus
from scenario.state import Relation, State

from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import ActiveStatus, BlockedStatus

from ..helpers import jsonpatch_delta, trigger


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        _call = None
        called = False

        def __init__(self, framework: Framework):
            super().__init__(framework)

            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if MyCharm._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_charm_heals_on_start(mycharm):
    def pre_event(charm):
        assert charm.unit.status == BlockedStatus('foo')
        assert not charm.called

    def call(charm, _):
        if charm.unit.status.message == 'foo':
            charm.unit.status = ActiveStatus('yabadoodle')

    def post_event(charm):
        assert charm.unit.status == ActiveStatus('yabadoodle')
        assert charm.called

    mycharm._call = call

    initial_state = State(
        config={'foo': 'bar'}, leader=True, unit_status=ScenarioBlockedStatus('foo')
    )

    out = trigger(
        initial_state,
        charm_type=mycharm,
        meta={'name': 'foo'},
        config={'options': {'foo': {'type': 'string'}}},
        event='start',
        post_event=post_event,
        pre_event=pre_event,
    )

    assert out.unit_status == ActiveStatus('yabadoodle')

    out_purged = dataclasses.replace(out, stored_states=initial_state.stored_states)
    assert jsonpatch_delta(out_purged, initial_state) == [
        {
            'op': 'replace',
            'path': '/unit_status/message',
            'value': 'yabadoodle',
        },
        {
            'op': 'replace',
            'path': '/unit_status/name',
            'value': 'active',
        },
    ]


def test_relation_data_access(mycharm):
    mycharm._call = lambda *_: True

    def check_relation_data(charm):
        foo_relations = charm.model.relations['relation_test']
        assert len(foo_relations) == 1
        foo_rel = foo_relations[0]
        assert len(foo_rel.units) == 2

        remote_units_data = {}
        for remote_unit in foo_rel.units:
            remote_units_data[remote_unit.name] = dict(foo_rel.data[remote_unit])

        remote_app_data = foo_rel.data[foo_rel.app]

        assert remote_units_data == {
            'karlos/0': {'foo': 'bar'},
            'karlos/1': {'baz': 'qux'},
        }

        assert remote_app_data == {'yaba': 'doodle'}

    state_in = State(
        relations={
            Relation(
                endpoint='relation_test',
                interface='azdrubales',
                id=1,
                remote_app_name='karlos',
                remote_app_data={'yaba': 'doodle'},
                remote_units_data={0: {'foo': 'bar'}, 1: {'baz': 'qux'}},
            )
        }
    )
    trigger(
        state_in,
        charm_type=mycharm,
        meta={
            'name': 'foo',
            'requires': {'relation_test': {'interface': 'azdrubales'}},
        },
        event='update_status',
        post_event=check_relation_data,
    )



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pytest
from scenario import Context
from scenario.state import State, _CharmSpec, _Event, _EventType

import ops


@pytest.mark.parametrize(
    'evt, expected_type',
    (
        ('foo_relation_changed', _EventType.RELATION),
        ('foo_relation_created', _EventType.RELATION),
        ('foo_bar_baz_relation_created', _EventType.RELATION),
        ('foo_storage_attached', _EventType.STORAGE),
        ('foo_storage_detaching', _EventType.STORAGE),
        ('foo_bar_baz_storage_detaching', _EventType.STORAGE),
        ('foo_pebble_ready', _EventType.WORKLOAD),
        ('foo_bar_baz_pebble_ready', _EventType.WORKLOAD),
        ('foo_pebble_custom_notice', _EventType.WORKLOAD),
        ('foo_bar_baz_pebble_custom_notice', _EventType.WORKLOAD),
        ('secret_remove', _EventType.SECRET),
        ('pre_commit', _EventType.FRAMEWORK),
        ('commit', _EventType.FRAMEWORK),
        ('collect_unit_status', _EventType.FRAMEWORK),
        ('collect_app_status', _EventType.FRAMEWORK),
        ('foo', _EventType.CUSTOM),
        ('kaboozle_bar_baz', _EventType.CUSTOM),
    ),
)
def test_event_type(evt, expected_type):
    event = _Event(evt)
    assert event._path.type is expected_type

    assert event._is_relation_event is (expected_type is _EventType.RELATION)
    assert event._is_storage_event is (expected_type is _EventType.STORAGE)
    assert event._is_workload_event is (expected_type is _EventType.WORKLOAD)
    assert event._is_secret_event is (expected_type is _EventType.SECRET)
    assert event._is_action_event is (expected_type is _EventType.ACTION)

    class MyCharm(ops.CharmBase):
        pass

    spec = _CharmSpec(
        MyCharm,
        meta={
            'requires': {
                'foo': {'interface': 'bar'},
                'foo_bar_baz': {'interface': 'bar'},
            },
            'storage': {
                'foo': {'type': 'filesystem'},
                'foo_bar_baz': {'type': 'filesystem'},
            },
            'containers': {'foo': {}, 'foo_bar_baz': {}},
        },
    )
    assert event._is_builtin_event(spec) is (expected_type is not _EventType.CUSTOM)


def test_emitted_framework():
    ctx = Context(ops.CharmBase, meta={'name': 'joop'}, capture_framework_events=True)
    ctx.run(ctx.on.update_status(), State())
    assert len(ctx.emitted_events) == 4
    assert list(map(type, ctx.emitted_events)) == [
        ops.UpdateStatusEvent,
        ops.CollectStatusEvent,
        ops.PreCommitEvent,
        ops.CommitEvent,
    ]


def test_emitted_deferred():
    class MyCharm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._on_update_status)

        def _on_update_status(self, _: ops.UpdateStatusEvent):
            pass

    ctx = Context(
        MyCharm,
        meta={'name': 'joop'},
        capture_deferred_events=True,
        capture_framework_events=True,
    )
    ctx.run(
        ctx.on.start(),
        State(deferred=[ctx.on.update_status().deferred(MyCharm._on_update_status)]),
    )

    assert len(ctx.emitted_events) == 5
    assert [e.handle.kind for e in ctx.emitted_events] == [
        'update_status',
        'start',
        'collect_unit_status',
        'pre_commit',
        'commit',
    ]



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark tests for ops-scenario."""

from __future__ import annotations

import dataclasses
import pathlib
import sys

import pytest

import ops
from ops import testing


@pytest.fixture
def benchmark_charm():
    charm_path = str(pathlib.Path(__file__).parent / 'charms' / 'benchmark_charm' / 'src')
    sys.path.append(charm_path)
    from benchmark_charm import BenchmarkCharm

    yield BenchmarkCharm
    sys.path.remove(charm_path)


# Note: the 'benchmark' argument here is a fixture that pytest-benchmark
# automatically makes available to all tests.
def test_context_explicit_meta(benchmark):
    ctx = benchmark(testing.Context, ops.CharmBase, meta={'name': 'foo'})
    assert isinstance(ctx, testing.Context)


def test_run_no_observer(benchmark, benchmark_charm):
    ctx = testing.Context(benchmark_charm)
    benchmark(ctx.run, ctx.on.start(), testing.State())
    # As with all these tests, we're interested in how long the
    # test took to run. However, we also verify that we did emit
    # an event (of a single kind across all the benchmark
    # repetitions) so that we are confident that something did
    # happen.
    assert len({e.handle.kind for e in ctx.emitted_events}) == 1


def test_run_observed(benchmark, benchmark_charm):
    ctx = testing.Context(benchmark_charm)
    benchmark(ctx.run, ctx.on.stop(), testing.State())
    assert len({e.handle.kind for e in ctx.emitted_events}) == 1


def test_context_explicit_meta_config_actions(benchmark):
    ctx = benchmark(
        testing.Context,
        ops.CharmBase,
        meta={'name': 'foo'},
        actions={'act': {'description': 'foo'}},
        config={'options': {'conf': {'type': 'int', 'description': 'bar'}}},
    )
    ctx.run(ctx.on.action('act'), testing.State(config={'conf': 10}))
    assert len({e.handle.kind for e in ctx.emitted_events}) == 1


def test_context_autoload_meta(benchmark, benchmark_charm):
    ctx = benchmark(testing.Context, benchmark_charm)
    assert isinstance(ctx, testing.Context)


def test_many_tests_explicit_meta(benchmark):
    # Calling benchmark(ctx.run, ...) benchmarks a test that looks like:
    # def test_x():
    #     ...
    #     ctx.run(ctx.on...)
    #     assert ...
    # In this test, we benchmark a test that looks like this instead:
    # def test_x():
    #     ...
    #     state = testing.State(...)
    #     state = ctx.run(ctx.on..., state)  # Event 1
    #     assert ...
    #     state = ctx.run(ctx.on..., state)  # Event 2
    #     assert ...
    # We generally recommend using a fresh context and having a single
    # run in each test function. However, there are times when it makes
    # sense to simulate multiple events in a row, and we want to make
    # sure that we are timing test that take this form as well, where the
    # cost of creating the Context object is amortised across many runs.

    def mock_pytest():
        """Simulate running multiple tests against the same charm."""
        ctx = testing.Context(ops.CharmBase, meta={'name': 'foo'})
        state = testing.State()
        for event in ('install', 'start', 'stop', 'remove'):
            state = ctx.run(getattr(ctx.on, event)(), state)

    benchmark(mock_pytest)


def test_many_tests_autoload_meta(benchmark, benchmark_charm):
    def mock_pytest():
        """Simulate running multiple tests against the same charm."""
        ctx = testing.Context(benchmark_charm)
        state = testing.State()
        for event in ('install', 'start', 'stop', 'remove'):
            state = ctx.run(getattr(ctx.on, event)(), state)

    benchmark(mock_pytest)


def test_lots_of_logs(benchmark, benchmark_charm):
    ctx = testing.Context(benchmark_charm)
    benchmark(ctx.run, ctx.on.update_status(), testing.State())
    assert len(ctx.juju_log) > 200


def test_full_state(benchmark, benchmark_charm):
    def fill_state():
        rel = testing.Relation('rel')
        peer = testing.PeerRelation('peer')
        network = testing.Network('MySpace')
        container = testing.Container('foo')
        storage = testing.Storage('bar')
        tcp = testing.TCPPort(22)
        icmp = testing.ICMPPort()
        udp = testing.UDPPort(8000)
        secret = testing.Secret({'password': 'admin'})
        resource = testing.Resource(name='baz', path='.')
        stored_state = testing.StoredState(owner_path='BenchMarkCharm')
        state = testing.State(
            relations={rel, peer},
            networks={network},
            containers={container},
            storages={storage},
            opened_ports={tcp, icmp, udp},
            secrets={secret},
            resources={resource},
            stored_states={stored_state},
            app_status=testing.ActiveStatus(),
            unit_status=testing.BlockedStatus("I'm stuck!"),
        )
        return state

    ctx = testing.Context(benchmark_charm)
    state_in = benchmark(fill_state)
    state_out = ctx.run(ctx.on.start(), state_in)
    # stored_states is complicated: it will contain a stored state the
    # framework itself added (counting the number of events), so the
    # input and output state doesn't naively match. We strip that out and
    # compare it separately.
    state_in_dict = dataclasses.asdict(state_in)
    state_out_dict = dataclasses.asdict(state_out)
    # An owner_path of None means that it's owned by the framework.
    assert state_in_dict['stored_states'] == {
        ss for ss in state_out_dict['stored_states'] if ss.owner_path is not None
    }
    del state_in_dict['stored_states']
    del state_out_dict['stored_states']
    assert state_in_dict == state_out_dict


def test_deferred_events(benchmark, benchmark_charm):
    ctx = testing.Context(benchmark_charm, capture_deferred_events=True)
    deferred = ctx.on.stop().deferred(benchmark_charm._on_stop)
    state_in = testing.State(deferred=[deferred])
    state_out = benchmark(ctx.run, ctx.on.config_changed(), state_in)
    assert len(state_out.deferred) == 1
    assert len({e.handle.kind for e in ctx.emitted_events}) == 2



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import pytest
from scenario import Context, State
from scenario.context import AlreadyEmittedError, Manager

from ops import ActiveStatus
from ops.charm import CharmBase, CollectStatusEvent


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        META: Mapping[str, Any] = {'name': 'mycharm'}
        ACTIONS: Mapping[str, Any] = {'do-x': {}}

        def __init__(self, framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, e):
            if isinstance(e, CollectStatusEvent):
                return

            self.unit.status = ActiveStatus(e.handle.kind)

    return MyCharm


def test_manager(mycharm):
    ctx = Context(mycharm, meta=mycharm.META)
    with Manager(ctx, ctx.on.start(), State()) as manager:
        assert isinstance(manager.charm, mycharm)
        state_out = manager.run()

    assert isinstance(state_out, State)


def test_manager_implicit(mycharm):
    ctx = Context(mycharm, meta=mycharm.META)
    with Manager(ctx, ctx.on.start(), State()) as manager:
        assert isinstance(manager.charm, mycharm)
        # do not call .run()

    # run is called automatically
    assert manager._emitted


def test_manager_reemit_fails(mycharm):
    ctx = Context(mycharm, meta=mycharm.META)
    with Manager(ctx, ctx.on.start(), State()) as manager:
        manager.run()
        with pytest.raises(AlreadyEmittedError):
            manager.run()


def test_context_manager(mycharm):
    ctx = Context(mycharm, meta=mycharm.META)
    with ctx(ctx.on.start(), State()) as manager:
        state_out = manager.run()
        assert isinstance(state_out, State)
    assert ctx.emitted_events[0].handle.kind == 'start'


def test_context_action_manager(mycharm):
    ctx = Context(mycharm, meta=mycharm.META, actions=mycharm.ACTIONS)
    with ctx(ctx.on.action('do-x'), State()) as manager:
        state_out = manager.run()
        assert isinstance(state_out, State)
    assert ctx.emitted_events[0].handle.kind == 'do_x_action'



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pytest
from scenario import Context
from scenario.errors import UncaughtCharmError
from scenario.state import (
    ActiveStatus,
    BlockedStatus,
    ErrorStatus,
    MaintenanceStatus,
    State,
    UnknownStatus,
    WaitingStatus,
)

import ops
from ops.charm import CharmBase
from ops.framework import Framework

from ..helpers import trigger


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_initial_status(mycharm):
    def post_event(charm: CharmBase):
        assert charm.unit.status == UnknownStatus()

    out = trigger(
        State(leader=True),
        'update_status',
        mycharm,
        meta={'name': 'local'},
        post_event=post_event,
    )

    assert out.unit_status == UnknownStatus()


def test_status_history(mycharm):
    class StatusCharm(mycharm):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._on_update_status)

        def _on_update_status(self, _):
            for obj in (self.unit, self.app):
                obj.status = ops.ActiveStatus('1')
                obj.status = ops.BlockedStatus('2')
                obj.status = ops.WaitingStatus('3')

    ctx = Context(
        StatusCharm,
        meta={'name': 'local'},
    )

    out = ctx.run(ctx.on.update_status(), State(leader=True))

    assert out.unit_status == WaitingStatus('3')
    assert ctx.unit_status_history == [
        UnknownStatus(),
        ActiveStatus('1'),
        BlockedStatus('2'),
    ]

    assert out.app_status == ops.WaitingStatus('3')
    assert ctx.app_status_history == [
        UnknownStatus(),
        ActiveStatus('1'),
        BlockedStatus('2'),
    ]


def test_status_history_preservation(mycharm):
    class StatusCharm(mycharm):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._on_update_status)

        def _on_update_status(self, _):
            for obj in (self.unit, self.app):
                obj.status = WaitingStatus('3')

    ctx = Context(
        StatusCharm,
        meta={'name': 'local'},
    )

    out = ctx.run(
        ctx.on.update_status(),
        State(
            leader=True,
            unit_status=ActiveStatus('foo'),
            app_status=ActiveStatus('bar'),
        ),
    )

    assert out.unit_status == WaitingStatus('3')
    assert ctx.unit_status_history == [ActiveStatus('foo')]

    assert out.app_status == WaitingStatus('3')
    assert ctx.app_status_history == [ActiveStatus('bar')]


def test_workload_history(mycharm):
    class WorkloadCharm(mycharm):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.install, self._on_install)
            framework.observe(self.on.start, self._on_start)
            framework.observe(self.on.update_status, self._on_update_status)

        def _on_install(self, _):
            self.unit.set_workload_version('1')

        def _on_start(self, _):
            self.unit.set_workload_version('1.1')

        def _on_update_status(self, _):
            self.unit.set_workload_version('1.2')

    ctx = Context(
        WorkloadCharm,
        meta={'name': 'local'},
    )

    out = ctx.run(ctx.on.install(), State(leader=True))
    out = ctx.run(ctx.on.start(), out)
    out = ctx.run(ctx.on.update_status(), out)

    assert ctx.workload_version_history == ['1', '1.1']
    assert out.workload_version == '1.2'


@pytest.mark.parametrize(
    'status',
    (
        ActiveStatus('foo'),
        WaitingStatus('bar'),
        BlockedStatus('baz'),
        MaintenanceStatus('qux'),
        ErrorStatus('fiz'),
        UnknownStatus(),
    ),
)
def test_status_comparison(status):
    if isinstance(status, UnknownStatus):
        ops_status = ops.UnknownStatus()
    else:
        ops_status = getattr(ops, status.__class__.__name__)(status.message)
    # A status can be compared to itself.
    assert status == status
    # A status can be compared to another instance of the scenario class.
    if isinstance(status, UnknownStatus):
        assert status == status.__class__()
    else:
        assert status == status.__class__(status.message)
    # A status can be compared to an instance of the ops class.
    assert status == ops_status
    # isinstance also works for comparing to the ops classes.
    assert isinstance(status, type(ops_status))
    # The repr of the scenario and ops classes should be identical.
    assert repr(status) == repr(ops_status)


@pytest.mark.parametrize(
    'status',
    (
        ActiveStatus('foo'),
        WaitingStatus('bar'),
        BlockedStatus('baz'),
        MaintenanceStatus('qux'),
    ),
)
def test_status_success(status: ops.StatusBase):
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._on_update_status)

        def _on_update_status(self, _):
            self.unit.status = status

    ctx = Context(MyCharm, meta={'name': 'foo'})
    ctx.run(ctx.on.update_status(), State())


@pytest.mark.parametrize(
    'status',
    (
        ErrorStatus('fiz'),
        UnknownStatus(),
    ),
)
def test_status_error(status: ops.StatusBase, monkeypatch: pytest.MonkeyPatch):
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._on_update_status)

        def _on_update_status(self, _):
            self.unit.status = status

    monkeypatch.setenv('SCENARIO_BARE_CHARM_ERRORS', 'false')
    ctx = Context(MyCharm, meta={'name': 'foo'})
    with pytest.raises(UncaughtCharmError) as excinfo:
        ctx.run(ctx.on.update_status(), State())
    assert isinstance(excinfo.value.__cause__, ops.ModelError)
    assert f'invalid status "{status.name}"' in str(excinfo.value.__cause__)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Test framework logger."""

from __future__ import annotations

import logging
import os

logger = logging.getLogger('ops-scenario')
logger.setLevel(os.getenv('OPS_SCENARIO_LOGGING', 'WARNING'))



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import dataclasses
import functools
import logging
import marshal
import re
import sys
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Generic, cast

import ops
import ops.storage
from ops._main import _Dispatcher, _Manager
from ops._main import logger as ops_logger
from ops.framework import _event_regex
from ops.log import JujuLogHandler, _get_juju_log_and_app_id

from .context import Context
from .errors import BadOwnerPath, NoObserverError
from .logger import logger as scenario_logger
from .mocking import _MockModelBackend
from .state import (
    CharmType,
    Container,
    DeferredEvent,
    RelationBase,
    Secret,
    Storage,
    StoredState,
)

if TYPE_CHECKING:  # pragma: no cover
    from .context import Context
    from .state import State, _CharmSpec, _Event

EVENT_REGEX = re.compile(_event_regex)
STORED_STATE_REGEX = re.compile(
    r'((?P<owner_path>.*)\/)?(?P<_data_type_name>\D+)\[(?P<name>.*)\]',
)

logger = scenario_logger.getChild('ops_main_mock')

# pyright: reportPrivateUsage=false


class UnitStateDB:
    """Wraps the unit-state database with convenience methods for adjusting the state."""

    def __init__(self, underlying_store: ops.storage.SQLiteStorage):
        self._db = underlying_store

    def get_stored_states(self) -> frozenset[StoredState]:
        """Load any StoredState data structures from the db."""
        db = self._db
        stored_states: set[StoredState] = set()
        for handle_path in db.list_snapshots():
            if not EVENT_REGEX.match(handle_path) and (
                match := STORED_STATE_REGEX.match(handle_path)
            ):
                stored_state_snapshot = db.load_snapshot(handle_path)
                kwargs = match.groupdict()
                sst = StoredState(content=stored_state_snapshot, **kwargs)
                stored_states.add(sst)

        return frozenset(stored_states)

    def get_deferred_events(self) -> list[DeferredEvent]:
        """Load any DeferredEvent data structures from the db."""
        db = self._db
        deferred: list[DeferredEvent] = []
        for handle_path in db.list_snapshots():
            if EVENT_REGEX.match(handle_path):
                notices = db.notices(handle_path)
                for handle, owner, observer in notices:
                    try:
                        snapshot_data = db.load_snapshot(handle)
                    except ops.storage.NoSnapshotError:
                        snapshot_data: dict[str, Any] = {}

                    event = DeferredEvent(
                        handle_path=handle,
                        owner=owner,
                        observer=observer,
                        snapshot_data=snapshot_data,
                    )
                    deferred.append(event)

        return deferred

    def apply_state(self, state: State):
        """Add DeferredEvent and StoredState from this State instance to the storage."""
        db = self._db
        for event in state.deferred:
            db.save_notice(event.handle_path, event.owner, event.observer)
            try:
                marshal.dumps(event.snapshot_data)
            except ValueError as e:
                raise ValueError(
                    f'unable to save the data for {event}, it must contain only simple types.',
                ) from e
            db.save_snapshot(event.handle_path, event.snapshot_data)

        for stored_state in state.stored_states:
            db.save_snapshot(stored_state._handle_path, stored_state.content)


class Ops(_Manager, Generic[CharmType]):
    """Class to manage stepping through ops setup, event emission and framework commit."""

    def __init__(
        self,
        state: State,
        event: _Event,
        context: Context[CharmType],
        charm_spec: _CharmSpec[CharmType],
        juju_context: ops.JujuContext,
    ):
        self.state = state
        self.event = event
        self.context = context
        self.charm_spec = charm_spec
        self.store = None
        self.captured_events: list[ops.EventBase] = []
        self.trace_data = []

        try:
            import ops_tracing._mock  # break circular import

            self._tracing_mock = ops_tracing._mock.patch_tracing()
            self._tracing_exporter = self._tracing_mock.__enter__()
        except ImportError:
            self._tracing_mock = None
            self._tracing_exporter = None

        model_backend = _MockModelBackend(
            state=state,
            event=event,
            context=context,
            charm_spec=charm_spec,
            juju_context=juju_context,
        )

        super().__init__(self.charm_spec.charm_type, model_backend, juju_context=juju_context)

    def _load_charm_meta(self):
        metadata = (self._charm_root / 'metadata.yaml').read_text()
        actions_meta = self._charm_root / 'actions.yaml'
        actions_metadata = actions_meta.read_text() if actions_meta.exists() else None
        config_meta = self._charm_root / 'config.yaml'
        config_metadata = config_meta.read_text() if config_meta.exists() else None

        return ops.CharmMeta.from_yaml(metadata, actions_metadata, config_metadata)

    @property
    def _framework_class(self):
        return functools.partial(CapturingFramework, context=self.context)

    def _setup_root_logging(self):
        # The warnings module captures this in _showwarning_orig, but we
        # shouldn't really be using a private method, so capture it ourselves as
        # well.
        original_showwarning = warnings.showwarning
        super()._setup_root_logging()
        # Ops also sets up logging to capture warnings, but we want the normal
        # output.
        logging.captureWarnings(False)
        warnings.showwarning = original_showwarning
        # Ops sets sys.excepthook to go to Juju's debug-log, but that's not
        # useful in a testing context, so we reset it here.
        sys.excepthook = sys.__excepthook__

    def _make_storage(self, _: _Dispatcher):
        # TODO: add use_juju_for_storage support
        storage = ops.storage.SQLiteStorage(':memory:')
        logger.info('Copying input state to storage.')
        self.store = UnitStateDB(storage)
        self.store.apply_state(self.state)
        return storage

    def _get_event_to_emit(self, event_name: str):
        if self.event._is_custom_event:
            assert self.event.custom_event is not None
            emitter_type = type(self.event.custom_event.emitter)
            # We require the event source to be an attribute on the charm object.
            # Essentially, we are replacing the unbound event with a bound one.
            for attr_name in dir(self.charm):
                attr = getattr(self.charm, attr_name)
                if not isinstance(attr, ops.Object):
                    continue
                for sub_attr_name in dir(attr):
                    sub_attr = getattr(attr, sub_attr_name)
                    if not isinstance(sub_attr, ops.ObjectEvents):
                        continue
                    if type(sub_attr) is not emitter_type:
                        continue
                    return getattr(sub_attr, self.event.custom_event.event_kind)

        owner = self._get_owner(self.charm, self.event.owner_path) if self.event else self.charm.on

        try:
            event_to_emit = getattr(owner, event_name)
        except AttributeError:
            ops_logger.debug('Event %s not defined for %s.', event_name, self.charm)
            raise NoObserverError(
                f'Cannot fire {event_name!r} on {owner}: invalid event (not on charm.on).',
            ) from None
        return event_to_emit

    def _object_to_ops_object(self, obj: Any) -> Any:
        if hasattr(obj, '_to_ops'):
            return obj._to_ops()
        if isinstance(obj, Secret):
            return self.charm.model.get_secret(id=obj.id, label=obj.label)
        if isinstance(obj, RelationBase):
            return self.charm.model.get_relation(obj.endpoint, obj.id)
        if isinstance(obj, Container):
            return self.charm.unit.get_container(obj.name)
        elif isinstance(obj, Storage):
            for storage in self.charm.model.storages[obj.name]:
                if storage.index == obj.index:
                    return storage
        return obj

    def _get_event_args(self, bound_event: ops.BoundEvent) -> tuple[list[Any], dict[str, Any]]:
        # For custom events, if the caller provided us with explicit args, we
        # merge them with the Juju ones (to handle libraries subclassing the
        # Juju events). We also handle converting from Scenario to ops types,
        # since the test code typically won't be able to create the ops objects,
        # as a model is required for many.
        args, kwargs = super()._get_event_args(bound_event)
        if self.event.custom_event_args is not None:
            for arg in self.event.custom_event_args:
                args.append(self._object_to_ops_object(arg))
        if self.event.custom_event_kwargs is not None:
            for key, value in self.event.custom_event_kwargs.items():
                kwargs[key] = self._object_to_ops_object(value)
        return args, kwargs

    @staticmethod
    def _get_owner(root: Any, path: Sequence[str]) -> ops.ObjectEvents:
        """Walk path on root to an ObjectEvents instance."""
        obj = root
        try:
            for step in path:
                obj = getattr(obj, step)
        except AttributeError:
            raise BadOwnerPath(
                f'event_owner_path {path!r} invalid: {step!r} leads to nowhere.',
            ) from None
        if not isinstance(obj, ops.ObjectEvents):
            raise BadOwnerPath(
                f'event_owner_path {path!r} invalid: does not lead to an ObjectEvents instance.',
            )
        return obj

    def _close(self):
        """Now that we're done processing this event, read the charm state and expose it."""
        logger.info('Copying storage to output state.')
        assert self.store is not None
        deferred = self.store.get_deferred_events()
        stored_state = self.store.get_stored_states()
        self.state = dataclasses.replace(self.state, deferred=deferred, stored_states=stored_state)

    def destroy(self):
        # Must be run first, so that ops._main trace span is finished and can be read back.
        super().destroy()
        if self._tracing_mock:
            assert self._tracing_exporter
            self.trace_data = self._tracing_exporter.get_finished_spans()
            self._tracing_mock.__exit__(None, None, None)

        # Clean up logging handlers to avoid test pollution.
        logger = logging.getLogger()
        for handler in logger.handlers[:]:
            if isinstance(handler, JujuLogHandler):
                logger.removeHandler(handler)
        # Clear app ID cached when logging security events.
        _get_juju_log_and_app_id.cache_clear()


class CapturingFramework(ops.Framework):
    def __init__(self, *args: Any, context: Context[CharmType], **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._context = context

    def _emit(self, event: ops.EventBase):
        if self._context.capture_framework_events or not isinstance(
            event,
            (ops.PreCommitEvent, ops.CommitEvent, ops.CollectStatusEvent),
        ):
            # Dump/undump the event to ensure any custom attributes are (re)set by restore().
            event.restore(event.snapshot())
            self._context.emitted_events.append(event)

        return super()._emit(event)

    def _reemit(self, single_event_path: str | None = None):
        if not self._context.capture_deferred_events or single_event_path is not None:
            return super()._reemit(single_event_path)

        # Load all notices from storage as events.
        for event_path, _, _ in self._storage.notices(single_event_path):
            event_handle = ops.Handle.from_path(event_path)
            try:
                event = self.load_snapshot(event_handle)
            except ops.NoTypeError:
                continue
            event = cast('ops.EventBase', event)
            event.deferred = False
            self._forget(event)  # prevent tracking conflicts

            # Dump/undump the event to ensure any custom attributes are (re)set by restore().
            event.restore(event.snapshot())
            self._context.emitted_events.append(event)

        return super()._reemit(single_event_path)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import dataclasses
import logging
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import jsonpatch
from scenario.context import _DEFAULT_JUJU_VERSION, Context
from scenario.state import _Event

if TYPE_CHECKING:  # pragma: no cover
    from scenario.state import CharmType, State

    _CT = TypeVar('_CT', bound=type[CharmType])

logger = logging.getLogger()


def trigger(
    state: State,
    event: str | _Event,
    charm_type: type[CharmType],
    pre_event: Callable[[CharmType], None] | None = None,
    post_event: Callable[[CharmType], None] | None = None,
    meta: dict[str, Any] | None = None,
    actions: dict[str, Any] | None = None,
    config: dict[str, Any] | None = None,
    charm_root: str | Path | None = None,
    juju_version: str = _DEFAULT_JUJU_VERSION,
) -> State:
    ctx = Context(
        charm_type=charm_type,
        meta=meta,
        actions=actions,
        config=config,
        charm_root=charm_root,
        juju_version=juju_version,
    )
    if isinstance(event, str):
        if event.startswith('relation_'):
            assert len(tuple(state.relations)) == 1, 'shortcut only works with one relation'
            event = getattr(ctx.on, event)(next(iter(state.relations)))
        elif event.startswith('pebble_'):
            assert len(tuple(state.containers)) == 1, 'shortcut only works with one container'
            event = getattr(ctx.on, event)(next(iter(state.containers)))
        else:
            event = getattr(ctx.on, event)()
    assert isinstance(event, _Event)
    with ctx(event, state=state) as mgr:
        if pre_event:
            pre_event(mgr.charm)
        state_out = mgr.run()
        if post_event:
            post_event(mgr.charm)
    return state_out


def jsonpatch_delta(self, other: State):
    dict_other = dataclasses.asdict(other)
    dict_self = dataclasses.asdict(self)
    for attr in (
        'relations',
        'containers',
        'storages',
        'opened_ports',
        'secrets',
        'resources',
        'stored_states',
        'networks',
    ):
        dict_other[attr] = [dataclasses.asdict(o) for o in dict_other[attr]]
        dict_self[attr] = [dataclasses.asdict(o) for o in dict_self[attr]]
    patch = jsonpatch.make_patch(dict_other, dict_self).patch
    return sort_patch(patch)


def sort_patch(patch: list[dict], key=lambda obj: obj['path'] + obj['op']):
    return sorted(patch, key=key)
